#!/usr/bin/python
#
# Copyright (c) 2020 Cole Neubauer, (@coleneubauer)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


ANSIBLE_METADATA = {'metadata_version': '1.1',
                    'status': ['preview'],
                    'supported_by': 'community'}


DOCUMENTATION = '''
---
module: azure_rm_backuppolicy
version_added: "1.4.0"
short_description: Manage Azure Backup Policy
description:
    - Create and delete instance of Azure Backup Policy.

options:
    vault_name:
        description:
            - The name of the Recovery Services Vault the policy belongs to.
        required: true
        type: str
    name:
        description:
            - The name of the backup policy.
        required: true
        type: str
    resource_group:
        description:
            - The name of the resource group the vault is in.
        required: true
        type: str
    state:
        description:
            - Assert the state of the backup policy.
            - Use C(present) to create or update a backup policy and C(absent) to delete it.
        default: present
        choices:
            - absent
            - present
        type: str
    backup_management_type:
        description:
            - Defines the type of resource the policy will be applied to.
        choices:
            - AzureIaasVM
        type: str
    policy_type:
        description:
            - Defines the type of backup policy
        default: V1
        choices:
            - V1
            - V2
        type: str
    schedule_run_time:
        description:
            - The hour to run backups.
            - Valid choices are on 24 hour scale (0-23).
        type: int
    instant_recovery_snapshot_retention:
        description:
            - How many days to retain instant recovery snapshots.
        type: int
    schedule_run_frequency:
        description:
            - The frequency to run the policy.
        choices:
            - Daily
            - Weekly
        type: str
    schedule_days:
        description:
            - List of days to execute the schedule.
            - Does not apply to Daily frequency.
        type: list
        elements: str
    weekly_retention_count:
        description:
            - The amount of weeks to retain backups.
        type: int
    daily_retention_count:
        description:
            - The amount of days to retain backups.
            - Does not apply to Weekly frequency.
        type: int
    schedule_weekly_frequency:
        description:
            - The amount of weeks between backups.
            - Backup every I(schedule_weekly_frequency) week(s).
            - Azure will default behavior to running weekly if this is left blank.
            - Does not apply to Daily frequency.
        type: int
    time_zone:
        description:
            - Timezone to apply I(schedule_run_time).
        default: UTC
        type: str
extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Cole Neubauer(@coleneubauer)
'''

EXAMPLES = '''
- name: Delete a backup policy
  azure_rm_backuppolicy:
    vault_name: Vault_Name
    name: Policy_Name
    resource_group: Resource_Group_Name
    state: absent

- name: Create a daily VM backup policy
  azure_rm_backuppolicy:
    vault_name: Vault_Name
    name: Policy_Name
    resource_group: Resource_Group_Name
    state: present
    backup_management_type: "AzureIaasVM"
    schedule_run_frequency: "Daily"
    instant_recovery_snapshot_retention: 2
    daily_retention_count: 12
    time_zone: "Pacific Standard Time"
    schedule_run_time: 14

- name: Create a weekly VM backup policy
  azure.azcollection.azure_rm_backuppolicy:
    vault_name: Vault_Name
    name: Policy_Name
    resource_group: Resource_Group_Name
    state: present
    backup_management_type: "AzureIaasVM"
    schedule_run_frequency: "Weekly"
    instant_recovery_snapshot_retention: 5
    weekly_retention_count: 4
    schedule_days:
      - "Monday"
      - "Wednesday"
      - "Friday"
    time_zone: "Pacific Standard Time"
    schedule_run_time: 8

- name: Create daily enhanced policy
  azure.azcollection.azure_rm_backuppolicy:
    vault_name: Vault_Name
    name: Policy_Name_Enhanced
    resource_group: Resource_Group_Name
    policy_type: V2
    state: present
    backup_management_type: AzureIaasVM
    schedule_run_frequency: Daily
    instant_recovery_snapshot_retention: 7
    daily_retention_count: 28
    time_zone: UTC
    schedule_run_time: 20
'''

RETURN = '''
id:
    description:
        - Id of specified backup policy.
    returned: always
    type: str
    sample: "/subscriptions/xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx/providers/Microsoft.RecoveryServices/vaults/Vault_Name/backupPolicies/Policy_Name"
location:
    description:
        - Location of backup policy.
    type: str
    returned: always
    sample: eastus
name:
    description:
        - Name of backup policy.
    type: str
    returned: always
    sample: DefaultPolicy
type:
    description:
        - Type of backup policy.
    type: str
    returned: always
    sample: Microsoft.RecoveryServices/vaults/backupPolicies
properties:
    description:
        - Attributes of the backup policy.
    type: dict
    returned: always
    sample: {
                "backup_management_type": "AzureIaasVM",
                "instant_rp_details": {},
                "instant_rp_retention_range_in_days": 5,
                "protected_items_count": 0,
                "retention_policy": {
                    "retention_policy_type": "LongTermRetentionPolicy",
                    "weekly_schedule": {
                        "days_of_the_week": [
                            "Monday",
                            "Wednesday",
                            "Thursday"
                        ],
                        "retention_duration": {
                            "count": 4,
                            "duration_type": "Weeks"
                        },
                        "retention_times": [
                            "2025-04-09T10:00:00.000Z"
                        ]
                    }
                },
                "schedule_policy": {
                    "schedule_policy_type": "SimpleSchedulePolicy",
                    "schedule_run_days": [
                        "Monday",
                        "Wednesday",
                        "Thursday"
                    ],
                    "schedule_run_frequency": "Weekly",
                    "schedule_run_times": [
                        "2025-04-09T10:00:00.000Z"
                    ],
                    "schedule_weekly_frequency": 0
                },
                "time_zone": "Pacific Standard Time"
            }
'''

from datetime import datetime
from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase

try:
    from azure.core.exceptions import ResourceNotFoundError

except ImportError:
    # This is handled in azure_rm_common
    pass


class AzureRMBackupPolicy(AzureRMModuleBase):
    """Configuration class for an Azure RM Backup Policy"""

    def __init__(self):
        self.module_arg_spec = dict(
            vault_name=dict(type='str', required=True),
            name=dict(type='str', required=True),
            resource_group=dict(type='str', required=True),
            state=dict(type='str', default='present', choices=['present', 'absent']),
            backup_management_type=dict(type='str', choices=['AzureIaasVM']),
            policy_type=dict(type='str', default='V1', choices=['V1', 'V2']),
            schedule_run_time=dict(type='int'),
            instant_recovery_snapshot_retention=dict(type='int'),
            schedule_run_frequency=dict(type='str', choices=['Daily', 'Weekly']),
            schedule_days=dict(type='list', elements='str'),
            weekly_retention_count=dict(type='int'),
            daily_retention_count=dict(type='int'),
            schedule_weekly_frequency=dict(type='int'),
            time_zone=dict(type='str', default='UTC'),
        )

        self.vault_name = None
        self.name = None
        self.resource_group = None
        self.backup_management_type = None
        self.policy_type = None
        self.schedule_run_time = None
        self.instant_recovery_snapshot_retention = None
        self.schedule_run_frequency = None
        self.schedule_days = None
        self.weekly_retention_count = None
        self.schedule_weekly_frequency = None
        self.daily_retention_count = None
        self.time_zone = None

        self.results = dict(
            changed=False,
            id=None,
        )

        required_if = [('schedule_run_frequency', 'Weekly', ['schedule_days', 'weekly_retention_count', 'schedule_run_time']),
                       ('schedule_run_frequency', 'Daily', ['daily_retention_count', 'schedule_run_time']),
                       ('state', 'present', ['schedule_run_frequency', 'backup_management_type']),
                       ('log_mode', 'file', ['log_path'])]

        super(AzureRMBackupPolicy, self).__init__(derived_arg_spec=self.module_arg_spec,
                                                  supports_check_mode=True,
                                                  supports_tags=False,
                                                  required_if=required_if)

    def exec_module(self, **kwargs):
        """Main module execution method"""

        for key in self.module_arg_spec:
            setattr(self, key, kwargs[key])

        existing_backup_policy = None
        response = None

        old_res = self.get_backup_policy()

        # either create or update
        if self.state == 'present':
            # check if the backup policy exists
            if not old_res:
                self.log("Backup policy {0} for vault {1} in resource group {2} does not exist.".format(self.name,
                                                                                                        self.vault_name,
                                                                                                        self.resource_group))

                self.results['changed'] = True

                if self.check_mode:
                    return self.results

                response = self.create_or_update_backup_policy()
                self.results.update(self.set_results(response))

            # log that we're doing an update
            else:
                self.log("Backup policy {0} for vault {1} in resource group {2} already exists, updating".format(self.name,
                                                                                                                 self.vault_name,
                                                                                                                 self.resource_group))
                if self.schedule_run_frequency is not None and\
                   self.schedule_run_frequency != old_res['properties']['schedule_policy']['schedule_run_frequency']:
                    self.results['changed'] = True
                elif self.time_zone is not None and self.time_zone != old_res['properties']['time_zone']:
                    self.results['changed'] = True
                elif self.instant_recovery_snapshot_retention is not None and\
                        self.instant_recovery_snapshot_retention != old_res['properties']['instant_rp_retention_range_in_days']:
                    self.results['changed'] = True
                elif self.schedule_weekly_frequency is not None and\
                        self.schedule_weekly_frequency != old_res['properties']['schedule_policy']['schedule_weekly_frequency']:
                    self.results['changed'] = True
                if self.weekly_retention_count is not None:
                    if self.weekly_retention_count !=\
                       old_res['properties']['retention_policy'].get('weekly_schedule', {}).get('retention_duration', {}).get('count'):
                        self.results['changed'] = True
                if self.daily_retention_count is not None:
                    if self.daily_retention_count !=\
                       old_res['properties']['retention_policy'].get('daily_schedule', {}).get('retention_duration', {}).get('count'):
                        self.results['changed'] = True
                if self.schedule_days is not None:
                    if old_res['properties']['schedule_policy'].get('schedule_run_days') is not None:
                        if set(self.schedule_days) != set(old_res['properties']['schedule_policy']['schedule_run_days']):
                            self.results['changed'] = True
                    else:
                        self.results['changed'] = True

                if self.check_mode:
                    self.results.update(old_res)
                    return self.results
                else:
                    if self.results['changed']:
                        response = self.create_or_update_backup_policy()
                        self.results.update(self.set_results(response))
                    else:
                        self.results.update(old_res)

        elif self.state == 'absent':
            if old_res:
                self.log("Delete backup policy")
                self.results['changed'] = True

                if self.check_mode:
                    self.results.update(old_res)
                    return self.results

                self.delete_backup_policy()

                self.log('backup policy deleted')

            else:
                # If backup policy doesn't exist, that's the desired state.
                self.log("Backup policy {0} for vault {1} in resource group {2} does not exist.".format(self.name,
                                                                                                        self.vault_name,
                                                                                                        self.resource_group))

        return self.results

    def create_or_update_backup_policy(self):
        '''
        Creates or updates backup policy.

        :return: ProtectionPolicyResource
        '''
        self.log("Creating backup policy {0} for vault {1} in resource group {2}".format(self.name,
                                                                                         self.vault_name,
                                                                                         self.resource_group))
        self.log("Creating backup policy in progress")

        response = None

        try:
            instant_rp_details = None
            # need to represent the run time as a date_time
            # year, month, day has no impact on run time but is more consistent to see it as the time of creation rather than hardcoded value
            dt = datetime.utcnow()
            dt = datetime(dt.year, dt.month, dt.day, 0, 0)

            # azure requires this as a list but at this time doesn't support multiple run times
            # should easily be converted at this step if they support it in the future
            schedule_run_times_as_datetimes = []
            schedule_run_time = self.schedule_run_time

            # basic parameter checking. try to provide a better description of faults than azure does at this time
            try:
                if 0 <= schedule_run_time <= 23:
                    schedule_run_times_as_datetimes = [(dt.replace(hour=schedule_run_time))]
                else:
                    raise ValueError('Paramater schedule_run_time {0} is badly formed must be on the 24 hour scale'.format(schedule_run_time))
                # azure forces instant_recovery_snapshot_retention to be 5 when schedule type is Weekly
                if self.schedule_run_frequency == "Weekly" and self.instant_recovery_snapshot_retention != 5:
                    raise ValueError('Paramater instant_recovery_snapshot_retention was {0} but must be 5 when schedule_run_frequency is Weekly'
                                     .format(self.instant_recovery_snapshot_retention))

                if self.schedule_run_frequency == "Weekly" and not (1 <= self.weekly_retention_count <= 5163):
                    raise ValueError('Paramater weekly_retention_count was {0} but must be between 1 and 5163 when schedule_run_frequency is Weekly'
                                     .format(self.weekly_retention_count))

                if self.schedule_run_frequency == "Daily" and not (7 <= self.daily_retention_count <= 9999):
                    raise ValueError('Paramater daily_retention_count was {0} but must be between 7 and 9999 when schedule_run_frequency is Daily'
                                     .format(self.daily_retention_count))

            except ValueError as e:
                self.results['changed'] = False
                self.fail(e)

            # create a schedule policy based on schedule_run_frequency
            if self.policy_type == 'V1':
                schedule_policy = self.recovery_services_backup_models.SimpleSchedulePolicy(schedule_run_frequency=self.schedule_run_frequency,
                                                                                            schedule_run_days=self.schedule_days,
                                                                                            schedule_run_times=schedule_run_times_as_datetimes,
                                                                                            schedule_weekly_frequency=self.schedule_weekly_frequency)
            elif self.policy_type == 'V2':
                schedule_policy = self.recovery_services_backup_models.SimpleSchedulePolicyV2(
                    schedule_run_frequency=self.schedule_run_frequency, daily_schedule=self.recovery_services_backup_models.WeeklySchedule(
                        schedule_run_days=self.schedule_days, schedule_run_times=schedule_run_times_as_datetimes))

            daily_retention_schedule = None
            weekly_retention_schedule = None

            # Daily backups can have a daily retention or weekly but Weekly backups cannot have a daily retention
            if (self.daily_retention_count and self.schedule_run_frequency == "Daily"):
                retention_duration = self.recovery_services_backup_models.RetentionDuration(count=self.daily_retention_count, duration_type="Days")
                daily_retention_schedule = self.recovery_services_backup_models.DailyRetentionSchedule(retention_times=schedule_run_times_as_datetimes,
                                                                                                       retention_duration=retention_duration)

            if (self.weekly_retention_count):
                retention_duration = self.recovery_services_backup_models.RetentionDuration(count=self.weekly_retention_count,
                                                                                            duration_type="Weeks")
                weekly_retention_schedule = self.recovery_services_backup_models.WeeklyRetentionSchedule(days_of_the_week=self.schedule_days,
                                                                                                         retention_times=schedule_run_times_as_datetimes,
                                                                                                         retention_duration=retention_duration)

            retention_policy = self.recovery_services_backup_models.LongTermRetentionPolicy(daily_schedule=daily_retention_schedule,
                                                                                            weekly_schedule=weekly_retention_schedule)

            policy_definition = None

            if self.backup_management_type == "AzureIaasVM":
                # This assignment exists exclusively to deal with the following line being too long otherwise
                AzureIaaSVMProtectionPolicy = self.recovery_services_backup_models.AzureIaaSVMProtectionPolicy
                policy_definition = AzureIaaSVMProtectionPolicy(instant_rp_details=instant_rp_details,
                                                                schedule_policy=schedule_policy,
                                                                retention_policy=retention_policy,
                                                                instant_rp_retention_range_in_days=self.instant_recovery_snapshot_retention,
                                                                time_zone=self.time_zone,
                                                                policy_type=self.policy_type)

            if policy_definition:
                policy_resource = self.recovery_services_backup_models.ProtectionPolicyResource(properties=policy_definition)
                response = self.recovery_services_backup_client.protection_policies.create_or_update(vault_name=self.vault_name,
                                                                                                     resource_group_name=self.resource_group,
                                                                                                     policy_name=self.name,
                                                                                                     parameters=policy_resource)

        except Exception as e:
            self.log('Error attempting to create the backup policy.')
            self.fail("Error creating the backup policy {0} for vault {1} in resource group {2}. Error Reads: {3}".format(self.name,
                                                                                                                          self.vault_name,
                                                                                                                          self.resource_group, e))

        return response

    def delete_backup_policy(self):
        '''
        Deletes specified backup policy.

        :return: ProtectionPolicyResource
        '''
        self.log("Deleting the backup policy {0} for vault {1} in resource group {2}".format(self.name, self.vault_name, self.resource_group))

        response = None

        try:
            response = self.recovery_services_backup_client.protection_policies.begin_delete(vault_name=self.vault_name,
                                                                                             resource_group_name=self.resource_group,
                                                                                             policy_name=self.name)

        except Exception as e:
            self.log('Error attempting to delete the backup policy.')
            self.fail("Error deleting the backup policy {0} for vault {1} in resource group {2}. Error Reads: {3}".format(self.name,
                                                                                                                          self.vault_name,
                                                                                                                          self.resource_group, e))

        return response

    def get_backup_policy(self):
        '''
        Gets the properties of the specified backup policy.

        :return: ProtectionPolicyResource
        '''
        self.log("Checking if the backup policy {0} for vault {1} in resource group {2} is present".format(self.name,
                                                                                                           self.vault_name,
                                                                                                           self.resource_group))

        policy = None

        try:
            policy = self.recovery_services_backup_client.protection_policies.get(vault_name=self.vault_name,
                                                                                  resource_group_name=self.resource_group,
                                                                                  policy_name=self.name)
        except ResourceNotFoundError as ex:
            self.log("Could not find backup policy {0} for vault {1} in resource group {2}".format(self.name, self.vault_name, self.resource_group))

        return self.set_results(policy)

    def set_results(self, policy):
        result = dict()
        if policy:
            result['id'] = policy.id
            result['location'] = policy.location
            result['name'] = policy.name
            result['type'] = policy.type
            result['properties'] = policy.properties.as_dict()
            return result
        else:
            return None


def main():
    """Main execution"""
    AzureRMBackupPolicy()


if __name__ == '__main__':
    main()



#
# (c) 2017 Red Hat Inc.
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import absolute_import, division, print_function


__metaclass__ = type

DOCUMENTATION = """
author:
 - Ansible Networking Team (@ansible-network)
name: default
short_description: Use default netconf plugin to run standard netconf commands as
  per RFC
description:
- This default plugin provides low level abstraction apis for sending and receiving
  netconf commands as per Netconf RFC specification.
version_added: 1.0.0
options:
  ncclient_device_handler:
    type: str
    default: default
    description:
    - Specifies the ncclient device handler name for network os that support default
      netconf implementation as per Netconf RFC specification. To identify the ncclient
      device handler name refer ncclient library documentation.
"""
import json

from ansible.module_utils.common.text.converters import to_text

from ansible_collections.ansible.netcommon.plugins.plugin_utils.netconf_base import NetconfBase


class Netconf(NetconfBase):
    def get_text(self, ele, tag):
        try:
            return to_text(ele.find(tag).text, errors="surrogate_then_replace").strip()
        except AttributeError:
            pass

    def get_device_info(self):
        device_info = dict()
        device_info["network_os"] = "default"
        return device_info

    def get_capabilities(self):
        result = dict()
        result["rpc"] = self.get_base_rpc()
        result["network_api"] = "netconf"
        result["device_info"] = self.get_device_info()
        result["server_capabilities"] = list(self.m.server_capabilities)
        result["client_capabilities"] = list(self.m.client_capabilities)
        result["session_id"] = self.m.session_id
        result["device_operations"] = self.get_device_operations(result["server_capabilities"])
        return json.dumps(result)



#!/usr/bin/python
# Copyright (c) 2021 Radek Sprta <mail@radeksprta.eu>
# Copyright (c) 2024 Colin Nolan <cn580@alumni.york.ac.uk>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: cargo
short_description: Manage Rust packages with cargo
version_added: 4.3.0
description:
  - Manage Rust packages with cargo.
author: "Radek Sprta (@radek-sprta)"
extends_documentation_fragment:
  - community.general._attributes
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
options:
  executable:
    description:
      - Path to the C(cargo) installed in the system.
      - If not specified, the module looks for C(cargo) in E(PATH).
    type: path
    version_added: 7.5.0
  name:
    description:
      - The name of a Rust package to install.
    type: list
    elements: str
    required: true
  path:
    description: The base path where to install the Rust packages. Cargo automatically appends V(/bin). In other words, V(/usr/local)
      becomes V(/usr/local/bin).
    type: path
  version:
    description: The version to install. If O(name) contains multiple values, the module tries to install all of them in this
      version.
    type: str
  locked:
    description:
      - Install with locked dependencies.
      - This is only used when installing packages.
    type: bool
    default: false
    version_added: 7.5.0
  state:
    description:
      - The state of the Rust package.
    type: str
    default: present
    choices: ["present", "absent", "latest"]
  directory:
    description:
      - Path to the source directory to install the Rust package from.
      - This is only used when installing packages.
    type: path
    version_added: 9.1.0
  features:
    description:
      - List of features to activate.
      - This is only used when installing packages.
    type: list
    elements: str
    default: []
    version_added: 11.0.0
requirements:
  - cargo installed
"""

EXAMPLES = r"""
- name: Install "ludusavi" Rust package
  community.general.cargo:
    name: ludusavi

- name: Install "ludusavi" Rust package with locked dependencies
  community.general.cargo:
    name: ludusavi
    locked: true

- name: Install "ludusavi" Rust package in version 0.10.0
  community.general.cargo:
    name: ludusavi
    version: '0.10.0'

- name: Install "ludusavi" Rust package to global location
  community.general.cargo:
    name: ludusavi
    path: /usr/local

- name: Remove "ludusavi" Rust package
  community.general.cargo:
    name: ludusavi
    state: absent

- name: Update "ludusavi" Rust package its latest version
  community.general.cargo:
    name: ludusavi
    state: latest

- name: Install "ludusavi" Rust package from source directory
  community.general.cargo:
    name: ludusavi
    directory: /path/to/ludusavi/source

- name: Install "serpl" Rust package with ast_grep feature
  community.general.cargo:
    name: serpl
    features:
      - ast_grep
"""

import json
import os
import re

from ansible.module_utils.basic import AnsibleModule


class Cargo:
    def __init__(self, module, **kwargs):
        self.module = module
        self.executable = [kwargs["executable"] or module.get_bin_path("cargo", True)]
        self.name = kwargs["name"]
        self.path = kwargs["path"]
        self.state = kwargs["state"]
        self.version = kwargs["version"]
        self.locked = kwargs["locked"]
        self.directory = kwargs["directory"]
        self.features = kwargs["features"]

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, path):
        if path is not None and not os.path.isdir(path):
            self.module.fail_json(msg=f"Path {path} is not a directory")
        self._path = path

    def _exec(self, args, run_in_check_mode=False, check_rc=True, add_package_name=True):
        if not self.module.check_mode or (self.module.check_mode and run_in_check_mode):
            cmd = self.executable + args
            rc, out, err = self.module.run_command(cmd, check_rc=check_rc)
            return out, err
        return "", ""

    def get_installed(self):
        cmd = ["install", "--list"]
        if self.path:
            cmd.append("--root")
            cmd.append(self.path)

        data, dummy = self._exec(cmd, True, False, False)

        package_regex = re.compile(r"^([\w\-]+) v(\S+).*:$")
        installed = {}
        for line in data.splitlines():
            package_info = package_regex.match(line)
            if package_info:
                installed[package_info.group(1)] = package_info.group(2)

        return installed

    def install(self, packages=None):
        cmd = ["install"]
        cmd.extend(packages or self.name)
        if self.locked:
            cmd.append("--locked")
        if self.path:
            cmd.append("--root")
            cmd.append(self.path)
        if self.version:
            cmd.append("--version")
            cmd.append(self.version)
        if self.directory:
            cmd.append("--path")
            cmd.append(self.directory)
        if self.features:
            cmd += ["--features", ",".join(self.features)]
        return self._exec(cmd)

    def is_outdated(self, name):
        installed_version = self.get_installed().get(name)
        latest_version = (
            self.get_latest_published_version(name) if not self.directory else self.get_source_directory_version(name)
        )
        return installed_version != latest_version

    def get_latest_published_version(self, name):
        cmd = ["search", name, "--limit", "1"]
        data, dummy = self._exec(cmd, True, False, False)

        match = re.search(r"^" + re.escape(name) + r'\s*=\s*"([^"]+)"', data, re.MULTILINE)
        if not match:
            self.module.fail_json(msg=f"No published version for package {name} found")
        return match.group(1)

    def get_source_directory_version(self, name):
        cmd = [
            "metadata",
            "--format-version",
            "1",
            "--no-deps",
            "--manifest-path",
            os.path.join(self.directory, "Cargo.toml"),
        ]
        data, dummy = self._exec(cmd, True, False, False)
        manifest = json.loads(data)

        package = next(
            (package for package in manifest["packages"] if package["name"] == name),
            None,
        )
        if not package:
            self.module.fail_json(
                msg=f"Package {name} not defined in source, found: {[x['name'] for x in manifest['packages']]}"
            )
        return package["version"]

    def uninstall(self, packages=None):
        cmd = ["uninstall"]
        cmd.extend(packages or self.name)
        return self._exec(cmd)


def main():
    arg_spec = dict(
        executable=dict(type="path"),
        name=dict(required=True, type="list", elements="str"),
        path=dict(type="path"),
        state=dict(default="present", choices=["present", "absent", "latest"]),
        version=dict(type="str"),
        locked=dict(default=False, type="bool"),
        directory=dict(type="path"),
        features=dict(default=[], type="list", elements="str"),
    )
    module = AnsibleModule(argument_spec=arg_spec, supports_check_mode=True)

    name = module.params["name"]
    state = module.params["state"]
    version = module.params["version"]
    directory = module.params["directory"]

    if not name:
        module.fail_json(msg="Package name must be specified")

    if directory is not None and not os.path.isdir(directory):
        module.fail_json(msg="Source directory does not exist")

    # Set LANG env since we parse stdout
    module.run_command_environ_update = dict(LANGUAGE="C", LC_ALL="C")

    cargo = Cargo(module, **module.params)
    changed, out, err = False, None, None
    installed_packages = cargo.get_installed()
    if state == "present":
        to_install = [
            n for n in name if (n not in installed_packages) or (version and version != installed_packages[n])
        ]
        if to_install:
            changed = True
            out, err = cargo.install(to_install)
    elif state == "latest":
        to_update = [n for n in name if n not in installed_packages or cargo.is_outdated(n)]
        if to_update:
            changed = True
            out, err = cargo.install(to_update)
    else:  # absent
        to_uninstall = [n for n in name if n in installed_packages]
        if to_uninstall:
            changed = True
            out, err = cargo.uninstall(to_uninstall)

    module.exit_json(changed=changed, stdout=out, stderr=err)


if __name__ == "__main__":
    main()



#!/usr/bin/python
#
# Copyright (c) 2019 Zim Kalinowski, (@zikalino)
#
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type


DOCUMENTATION = '''
---
module: azure_rm_managementgroup
version_added: '0.1.1'
short_description: Manage Azure ManagementGroup instance
description:
    - Create, update and delete instance of Azure ManagementGroup.
options:
    group_id:
        description:
            - Management Group ID.
        type: str
        required: true
    name:
        description:
            - The name of the management group.
            - For example 00000000-0000-0000-0000-000000000000.
        type: str
    id:
        description:
            - The fully qualified ID for the management group.
            - For example /providers/Microsoft.Management/managementGroups/0000000-0000-0000-0000-000000000000.
        type: str
    type:
        description:
            - The type of the resource.
            - For example /providers/Microsoft.Management/managementGroups.
        type: str
    properties:
        description:
            - The properties of the management group.
        type: dict
        suboptions:
            tenant_id:
                description:
                    - The AAD Tenant ID associated with the management group.
                    - For example 00000000-0000-0000-0000-000000000000.
                type: str
            display_name:
                description:
                    - The friendly name of the management group.
                type: str
            parent_id:
                description:
                    - The parent of the management group.
                type: str
    state:
        description:
            - Assert the state of the ManagementGroup.
            - Use C(present) to create or update an ManagementGroup and C(absent) to delete it.
        default: present
        choices:
            - absent
            - present
        type: str
extends_documentation_fragment:
    - azure.azcollection.azure
    - azure.azcollection.azure_tags
author:
    - Zim Kalinowski (@zikalino)

'''

EXAMPLES = '''
- name: Create new management group
  azure_rm_managementgroup:
    group_id: test
    type: /providers/Microsoft.Management/managementGroups/
    name: test

- name: create management group
  azure_rm_managementgroup:
    group_id: ChildGroup
    type: /providers/Microsoft.Management/managementGroups/
    name: ChildGroup
    properties:
      tenant_id: 20000000-0000-0000-0000-000000000000
      display_name: ChildGroup
      parent_id: /providers/Microsoft.Management/managementGroups/RootGroup

- name: delete management group
  azure_rm_managementgroup:
    group_id: ChildGroup
    state: absent
'''

RETURN = '''
id:
    description:
        - The fully qualified ID for the management group.
        - For example /providers/Microsoft.Management/managementGroups/0000000-0000-0000-0000-000000000000.
    returned: always
    type: str
type:
    description:
        - The type of the resource.
        - For example /providers/Microsoft.Management/managementGroups.
    returned: always
    type: str
name:
    description:
        - The name of the management group.
        - For example 00000000-0000-0000-0000-000000000000.
    returned: always
    type: str
properties:
    description:
        - The properties of the management group.
    returned: always
    type: complex
    contains:
        tenant_id:
            description:
                - The AAD Tenant ID associated with the management group.
                - For example 00000000-0000-0000-0000-000000000000.
            returned: always
            type: str
        display_name:
            description:
                - The friendly name of the management group.
            returned: always
            type: str
        roles:
            description:
                - The role definitions associated with the management group.
            returned: always
            type: str
        details:
            description:
                - The details of the management group.
            returned: always
            type: complex
            contains:
                version:
                    description:
                        - The version number of the object.
                    returned: always
                    type: str
                updated_time:
                    description:
                        - The date and time when this object was last updated.
                    returned: always
                    type: str
                updated_by:
                    description:
                        - The identity of the principal or process that updated the object.
                    returned: always
                    type: str
                parent:
                    description:
                        - The parent of the management group.
                    returned: always
                    type: complex
                    contains:
                        id:
                            description:
                                - The fully qualified ID for the parent management group.
                                - For example /providers/Microsoft.Management/managementGroups/0000000-0000-0000-0000-000000000000.
                            returned: always
                            type: str
                        name:
                            description:
                                - The name of the parent management group
                            returned: always
                            type: str
                        display_name:
                            description:
                                - The friendly name of the parent management group.
                            returned: always
                            type: str
                children:
                    description:
                        - The list of children.
                    returned: always
                    type: complex
                    contains:
                        type:
                            description:
                                - The fully qualified resource type which includes provider namespace.
                                - For example /providers/Microsoft.Management/managementGroups.
                            returned: always
                            type: str
                        id:
                            description:
                                - The fully qualified ID for the child resource (management group or subscription).
                                - For example /providers/Microsoft.Management/managementGroups/0000000-0000-0000-0000-000000000000.
                            returned: always
                            type: str
                        name:
                            description:
                                - The name of the child entity.
                            returned: always
                            type: str
                        display_name:
                            description:
                                - The friendly name of the child resource.
                            returned: always
                            type: str

'''

import time
import json
from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_rest import GenericRestClient
from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common_ext import AzureRMModuleBaseExt


class Actions:
    NoAction, Create, Update, Delete = range(4)


class AzureRMManagementGroups(AzureRMModuleBaseExt):
    def __init__(self):
        self.module_arg_spec = dict(
            group_id=dict(type='str', required=True),
            name=dict(type='str'),
            id=dict(type='str'),
            type=dict(type='str'),
            properties=dict(
                type='dict',
                options=dict(
                    tenant_id=dict(type='str'),
                    display_name=dict(type='str'),
                    parent_id=dict(type='str')
                )
            ),
            state=dict(type='str', default='present', choices=['present', 'absent']),
        )

        self.group_id = None
        self.state = None

        self.results = dict(changed=False)
        self.mgmt_client = None
        self.url = None
        self.status_code = [200, 201, 202]
        self.to_do = Actions.NoAction
        self.body = {}
        self.query_parameters = {}
        self.query_parameters['api-version'] = '2018-03-01-preview'
        self.header_parameters = {}
        self.header_parameters['Content-Type'] = 'application/json; charset=utf-8'

        super(AzureRMManagementGroups, self).__init__(derived_arg_spec=self.module_arg_spec,
                                                      supports_check_mode=True,
                                                      supports_tags=True)

    def exec_module(self, **kwargs):
        for key in list(self.module_arg_spec.keys()):
            if hasattr(self, key):
                setattr(self, key, kwargs[key])
            elif key == 'properties' and kwargs[key] is not None:
                self.body['properties'] = {}
                for item in kwargs['properties'].keys():
                    if item == 'tenant_id':
                        self.body['properties']['tenantId'] = kwargs['properties'][item]
                    elif item == 'display_name':
                        self.body['properties']['displayName'] = kwargs['properties'][item]
                    elif item == 'parent_id':
                        self.body['properties']['details'] = {}
                        self.body['properties']['details']['parent'] = {}
                        self.body['properties']['details']['parent']['id'] = kwargs['properties'][item]

            elif kwargs[key] is not None:
                self.body[key] = kwargs[key]

        old_response = None
        response = None

        self.mgmt_client = self.get_mgmt_svc_client(GenericRestClient,
                                                    base_url=self._cloud_environment.endpoints.resource_manager)

        self.url = ('/providers' +
                    '/Microsoft.Management' +
                    '/managementGroups' +
                    '/{{ management_group_name }}')
        self.url = self.url.replace('{{ management_group_name }}', self.group_id)

        old_response = self.get_resource()

        if not old_response:
            self.log("ManagementGroup instance doesn't exist")

            if self.state == 'absent':
                self.log("Old instance didn't exist")
            else:
                self.to_do = Actions.Create
        else:
            self.log('ManagementGroup instance already exists')

            if self.state == 'absent':
                self.to_do = Actions.Delete
            else:
                for key in self.body.keys():
                    if key == 'properties':
                        if old_response.get('properties') is None or \
                           not all(self.body['properties'][item] == old_response['properties'].get(item)
                           for item in self.body['properties'].keys()):
                            self.to_do = Actions.Update
                    elif self.body[key] != old_response.get(key):
                        self.to_do = Actions.Update

        if (self.to_do == Actions.Create) or (self.to_do == Actions.Update):
            self.log('Need to Create / Update the ManagementGroup instance')

            if self.check_mode:
                self.results['changed'] = True
                return self.results

            response = self.create_update_resource()

            self.results['body'] = self.body
            # if not old_response:
            self.results['changed'] = True
            # else:
            #     self.results['changed'] = old_response.__ne__(response)
            self.log('Creation / Update done')
        elif self.to_do == Actions.Delete:
            self.log('ManagementGroup instance deleted')
            self.results['changed'] = True

            if self.check_mode:
                return self.results

            self.delete_resource()

            # make sure instance is actually deleted, for some Azure resources, instance is hanging around
            # for some time after deletion -- this should be really fixed in Azure
            while self.get_resource():
                time.sleep(20)
        else:
            self.log('ManagementGroup instance unchanged')
            self.results['changed'] = False
            response = old_response

        if response:
            self.results["id"] = response["id"]
            self.results["type"] = response["type"]
            self.results["name"] = response["name"]
            self.results["properties"] = response["properties"]

        return self.results

    def create_update_resource(self):
        # self.log('Creating / Updating the ManagementGroup instance {0}'.format(self.))

        try:
            if self.to_do == Actions.Create:
                response = self.mgmt_client.query(self.url,
                                                  'PUT',
                                                  self.query_parameters,
                                                  self.header_parameters,
                                                  self.body,
                                                  self.status_code,
                                                  600,
                                                  30)
            else:
                response = self.mgmt_client.query(self.url,
                                                  'PATCH',
                                                  self.query_parameters,
                                                  self.header_parameters,
                                                  self.body,
                                                  self.status_code,
                                                  600,
                                                  30)
        except Exception as exc:
            self.log('Error attempting to create the ManagementGroup instance.')
            self.fail('Error creating the ManagementGroup instance: {0}'.format(str(exc)))

        if hasattr(response, 'body'):
            response = json.loads(response.body())
        elif hasattr(response, 'context'):
            response = response.context['deserialized_data']
        else:
            self.fail("Create or Updating fail, no match message return, return info as {0}".format(response))

        return response

    def delete_resource(self):
        # self.log('Deleting the ManagementGroup instance {0}'.format(self.))
        try:
            response = self.mgmt_client.query(self.url,
                                              'DELETE',
                                              self.query_parameters,
                                              self.header_parameters,
                                              None,
                                              self.status_code,
                                              600,
                                              30)
        except Exception as e:
            self.log('Error attempting to delete the ManagementGroup instance.')
            self.fail('Error deleting the ManagementGroup instance: {0}'.format(str(e)))

        return True

    def get_resource(self):
        # self.log('Checking if the ManagementGroup instance {0} is present'.format(self.))
        found = False
        try:
            response = self.mgmt_client.query(self.url,
                                              'GET',
                                              self.query_parameters,
                                              self.header_parameters,
                                              None,
                                              self.status_code,
                                              600,
                                              30)
            found = True
            response = json.loads(response.body())
            self.log("Response : {0}".format(response))
            # self.log("ManagementGroup instance : {0} found".format(response.name))
        except Exception as e:
            self.log('Did not find the ManagementGroup instance. msg: {0}'.format(e))
        if found is True:
            return response

        return False


def main():
    AzureRMManagementGroups()


if __name__ == '__main__':
    main()



#!/usr/bin/python
# Copyright (c) 2016 Michael Gruener <michael.gruener@chaosmoon.net>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: acme_certificate
author: "Michael Gruener (@mgruener)"
short_description: Create SSL/TLS certificates with the ACME protocol
description:
  - Create and renew SSL/TLS certificates with a CA supporting the L(ACME protocol,https://tools.ietf.org/html/rfc8555), such
    as L(Let's Encrypt,https://letsencrypt.org/).
    The current implementation supports the V(http-01), V(dns-01), V(dns-account-01), V(dns-persist-01), and V(tls-alpn-01) challenges.
  - To use this module, it has to be executed twice. Either as two different tasks in the same run or during two runs. Note
    that the output of the first run needs to be recorded and passed to the second run as the module argument O(data).
  - Between these two tasks you have to fulfill the required steps for the chosen challenge by whatever means necessary.
    For V(http-01) that means creating the necessary challenge file on the destination webserver.
    For V(dns-01), V(dns-account-01), and V(dns-persist-01) the necessary DNS records have to be created.
    For V(tls-alpn-01) the necessary certificate has to be created and served.
    It is I(not) the responsibility of this module to perform these steps.
  - For details on how to fulfill these challenges, you might have to read through L(the main ACME specification,https://tools.ietf.org/html/rfc8555#section-8)
    and the L(TLS-ALPN-01 specification,https://www.rfc-editor.org/rfc/rfc8737.html#section-3). Also, consider the examples
    provided for this module.
  - The module support for IP identifiers according to L(RFC 8738,https://www.rfc-editor.org/rfc/rfc8738.html).
  - The module B(experimentally) supports the V(dns-account-01) challenge type according to
    L(acme-dns-account-label draft 02, https://datatracker.ietf.org/doc/html/draft-ietf-acme-dns-account-label-02).
    Note that the supported draft version can change at any time,
    and changes will only be considered breaking once the draft reached RFC status.
  - The module B(experimentally) supports the V(dns-persist-01) challenge type according to
    L(acme-dns-persist draft 01, https://www.ietf.org/archive/id/draft-ietf-acme-dns-persist-01.html).
    Note that the supported draft version can change at any time,
    and changes will only be considered breaking once the draft reached RFC status.
notes:
  - At least one of O(dest) and O(fullchain_dest) must be specified.
  - This module includes basic account management functionality. If you want to have more control over your ACME account,
    use the M(community.crypto.acme_account) module and disable account management for this module using the O(modify_account)
    option.
  - This module was called C(letsencrypt) before Ansible 2.6. The usage did not change.
seealso:
  - name: The Let's Encrypt documentation
    description: Documentation for the Let's Encrypt Certification Authority. Provides useful information for example on rate
      limits.
    link: https://letsencrypt.org/docs/
  - name: Automatic Certificate Management Environment (ACME)
    description: The specification of the ACME protocol (RFC 8555).
    link: https://tools.ietf.org/html/rfc8555
  - name: ACME TLS ALPN Challenge Extension
    description: The specification of the V(tls-alpn-01) challenge (RFC 8737).
    link: https://www.rfc-editor.org/rfc/rfc8737.html
  - module: community.crypto.acme_challenge_cert_helper
    description: Helps preparing V(tls-alpn-01) challenges.
  - module: community.crypto.openssl_privatekey
    description: Can be used to create private keys (both for certificates and accounts).
  - module: community.crypto.openssl_privatekey_pipe
    description: Can be used to create private keys without writing it to disk (both for certificates and accounts).
  - module: community.crypto.openssl_csr
    description: Can be used to create a Certificate Signing Request (CSR).
  - module: community.crypto.openssl_csr_pipe
    description: Can be used to create a Certificate Signing Request (CSR) without writing it to disk.
  - module: community.crypto.certificate_complete_chain
    description: Allows to find the root certificate for the returned fullchain.
  - module: community.crypto.acme_certificate_revoke
    description: Allows to revoke certificates.
  - module: community.crypto.acme_account
    description: Allows to create, modify or delete an ACME account.
  - module: community.crypto.acme_inspect
    description: Allows to debug problems.
  - module: community.crypto.acme_certificate_deactivate_authz
    description: Allows to deactivate (invalidate) ACME v2 orders.
extends_documentation_fragment:
  - community.crypto._acme.basic
  - community.crypto._acme.account
  - community.crypto._acme.certificate
  - community.crypto._attributes
  - community.crypto._attributes.files
  - community.crypto._attributes.actiongroup_acme
attributes:
  check_mode:
    support: full
  diff_mode:
    support: none
  safe_file_operations:
    support: full
  idempotent:
    support: partial
    details:
      - If O(force=true), the module is not idempotent.
        If O(force=false), it depends on the certificate's validity period and the value of O(remaining_days).
      - The second phase invocation of the module is always idempotent, assuming no error occurs.
options:
  account_email:
    description:
      - The email address associated with this account.
      - It will be used for certificate expiration warnings.
      - Note that when O(modify_account) is not set to V(false) and you also used the M(community.crypto.acme_account) module
        to specify more than one contact for your account, this module will update your account and restrict it to the (at
        most one) contact email address specified here.
    type: str
  agreement:
    description:
      - URI to a terms of service document you agree to when using the ACME v1 service at O(acme_directory).
      - Default is latest gathered from O(acme_directory) URL.
      - This option has no longer any effect. It is deprecated and will be removed from community.crypto 4.0.0.
    type: str
  terms_agreed:
    description:
      - Boolean indicating whether you agree to the terms of service document.
      - ACME servers can require this to be true.
    type: bool
    default: false
  modify_account:
    description:
      - Boolean indicating whether the module should create the account if necessary, and update its contact data.
      - Set to V(false) if you want to use the M(community.crypto.acme_account) module to manage your account instead, and
        to avoid accidental creation of a new account using an old key if you changed the account key with M(community.crypto.acme_account).
      - If set to V(false), O(terms_agreed) and O(account_email) are ignored.
      - The current default V(true) is B(deprecated) and will change to V(false) in community.crypto 4.0.0.
    type: bool
  challenge:
    description:
      - The challenge to be performed.
      - If set to V(no challenge), no challenge will be used. This is necessary for some private CAs which use External Account
        Binding and other means of validating certificate assurance. For example, an account could be allowed to issue certificates
        for C(foo.example.com) without any further validation for a certain period of time.
      - Support for V(dns-account-01) and V(dns-persist-01) has been added in community.crypto 3.2.0.
    type: str
    default: http-01
    choices:
      - http-01
      - dns-01
      - dns-account-01
      - dns-persist-01
      - tls-alpn-01
      - no challenge
  csr:
    aliases: ['src']
  csr_content:
    version_added: 1.2.0
  data:
    description:
      - The data to validate ongoing challenges. This must be specified for the second run of the module only.
      - The value that must be used here will be provided by a previous use of this module. See the examples for more details.
      - Note that for ACME v2, only the C(order_uri) entry of O(data) will be used. For ACME v1, O(data) must be non-empty
        to indicate the second stage is active; all needed data will be taken from the CSR.
      - 'I(Note): the O(data) option was marked as C(no_log) up to Ansible 2.5. From Ansible 2.6 on, it is no longer marked
        this way as it causes error messages to be come unusable, and O(data) does not contain any information which can be
        used without having access to the account key or which are not public anyway.'
    type: dict
  dest:
    description:
      - The destination file for the certificate.
      - Required if O(fullchain_dest) is not specified.
    type: path
    aliases: ['cert']
  fullchain_dest:
    description:
      - The destination file for the full chain (that is, a certificate followed by chain of intermediate certificates).
      - Required if O(dest) is not specified.
    type: path
    aliases: ['fullchain']
  chain_dest:
    description:
      - If specified, the intermediate certificate will be written to this file.
    type: path
    aliases: ['chain']
  remaining_days:
    description:
      - The number of days the certificate must have left being valid. If RV(cert_days) < O(remaining_days), then it will
        be renewed. If the certificate is not renewed, module return values will not include RV(challenge_data).
      - To make sure that the certificate is renewed in any case, you can use the O(force) option.
    type: int
    default: 10
  deactivate_authzs:
    description:
      - Deactivate authentication objects (authz) after issuing a certificate, or when issuing the certificate failed.
      - Authentication objects are bound to an account key and remain valid for a certain amount of time, and can be used
        to issue certificates without having to re-authenticate the domain. This can be a security concern.
    type: bool
    default: false
  force:
    description:
      - Enforces the execution of the challenge and validation, even if an existing certificate is still valid for more than
        O(remaining_days).
      - This is especially helpful when having an updated CSR, for example with additional domains for which a new certificate
        is desired.
    type: bool
    default: false
  retrieve_all_alternates:
    description:
      - When set to V(true), will retrieve all alternate trust chains offered by the ACME CA. These will not be written to
        disk, but will be returned together with the main chain as RV(all_chains). See the documentation for the RV(all_chains)
        return value for details.
    type: bool
    default: false
  select_chain:
    description:
      - Allows to specify criteria by which an (alternate) trust chain can be selected.
      - The list of criteria will be processed one by one until a chain is found matching a criterium. If such a chain is
        found, it will be used by the module instead of the default chain.
      - If a criterium matches multiple chains, the first one matching will be returned. The order is determined by the ordering
        of the C(Link) headers returned by the ACME server and might not be deterministic.
      - Every criterium can consist of multiple different conditions, like O(select_chain[].issuer) and O(select_chain[].subject).
        For the criterium to match a chain, all conditions must apply to the same certificate in the chain.
      - This option can only be used with the C(cryptography) backend.
    type: list
    elements: dict
    version_added: '1.0.0'
    suboptions:
      test_certificates:
        description:
          - Determines which certificates in the chain will be tested.
          - V(all) tests all certificates in the chain (excluding the leaf, which is identical in all chains).
          - V(first) only tests the first certificate in the chain, that is the one which signed the leaf.
          - V(last) only tests the last certificate in the chain, that is the one furthest away from the leaf. Its issuer
            is the root certificate of this chain.
        type: str
        default: all
        choices: [first, last, all]
      issuer:
        description:
          - Allows to specify parts of the issuer of a certificate in the chain must have to be selected.
          - If O(select_chain[].issuer) is empty, any certificate will match.
          - 'An example value would be V({"commonName": "My Preferred CA Root"}).'
        type: dict
      subject:
        description:
          - Allows to specify parts of the subject of a certificate in the chain must have to be selected.
          - If O(select_chain[].subject) is empty, any certificate will match.
          - 'An example value would be V({"CN": "My Preferred CA Intermediate"}).'
        type: dict
      subject_key_identifier:
        description:
          - Checks for the SubjectKeyIdentifier extension. This is an identifier based on the private key of the intermediate
            certificate.
          - The identifier must be of the form V(A8:4A:6A:63:04:7D:DD:BA:E6:D1:39:B7:A6:45:65:EF:F3:A8:EC:A1).
        type: str
      authority_key_identifier:
        description:
          - Checks for the AuthorityKeyIdentifier extension. This is an identifier based on the private key of the issuer
            of the intermediate certificate.
          - The identifier must be of the form V(C4:A7:B1:A4:7B:2C:71:FA:DB:E1:4B:90:75:FF:C4:15:60:85:89:10).
        type: str
  include_renewal_cert_id:
    description:
      - Determines whether to request renewal of an existing certificate according to L(Section 5 of RFC 9773,
        https://www.rfc-editor.org/rfc/rfc9773.html#section-5).
      - This is only used when the certificate specified in O(dest) or O(fullchain_dest) already exists.
      - Generally you should use V(when_ari_supported) if you know that the ACME service supports a compatible draft (or final
        version, once it is out) of the ARI extension. V(always) should never be necessary. If you are not sure, or if you
        receive strange errors on invalid C(replaces) values in order objects, use V(never), which also happens to be the
        default.
      - ACME servers might refuse to create new orders with C(replaces) for certificates that already have an existing order.
        This can happen if this module is used to create an order, and then the playbook/role fails in case the challenges
        cannot be set up. If the playbook/role does not record the order data to continue with the existing order, but tries
        to create a new one on the next run, creating the new order might fail. If O(order_creation_error_strategy=fail)
        this will make the module fail. O(order_creation_error_strategy=auto) and
        O(order_creation_error_strategy=retry_without_replaces_cert_id) will avoid this by leaving away C(replaces)
        on retries.
      - If O(order_creation_error_strategy=fail), for the above reason, this option should only be set to a value different
        from V(never) if the role/playbook using it keeps track of order data accross restarts, or if it takes care to
        deactivate orders whose processing is aborted. Orders can be deactivated with the
        M(community.crypto.acme_certificate_deactivate_authz) module.
    type: str
    choices:
      never: Never send the certificate ID of the certificate to renew.
      when_ari_supported: Only send the certificate ID if the ARI endpoint is found in the ACME directory.
      always: Will always send the certificate ID of the certificate to renew.
    default: never
    version_added: 2.20.0
  profile:
    description:
      - Chose a specific profile for certificate selection. The available profiles depend on the CA.
      - See L(a blog post by Let's Encrypt, https://letsencrypt.org/2025/01/09/acme-profiles/) and
        L(draft-aaron-acme-profiles-01, https://datatracker.ietf.org/doc/draft-aaron-acme-profiles/)
        for more information.
    type: str
    version_added: 2.24.0
  order_creation_error_strategy:
    description:
      - Selects the error handling strategy for ACME protocol errors if creating a new ACME order fails.
    type: str
    choices:
      auto:
        - An unspecified algorithm that tries to be clever.
        - Right now identical to V(retry_without_replaces_cert_id).
      always:
        - Always retry, until the limit in O(order_creation_max_retries) has been reached.
      fail:
        - Simply fail in case of errors. Do not attempt to retry.
        - This has been the default before community.crypto 2.24.0.
      retry_without_replaces_cert_id:
        - If O(include_renewal_cert_id) is present, creating the order will be tried again without C(replaces).
        - The only exception is an error of type C(urn:ietf:params:acme:error:alreadyReplaced), that indicates that
          the certificate was already replaced. This usually means something went wrong and the user should investigate.
    default: auto
    version_added: 2.24.0
  order_creation_max_retries:
    description:
      - Depending on the strategy selected in O(order_creation_error_strategy), will retry creating new orders
        for at most the specified amount of times.
    type: int
    default: 3
    version_added: 2.24.0
"""

EXAMPLES = r"""
---
### Example with HTTP challenge ###

- name: Create a challenge for sample.com using a account key from a variable.
  community.crypto.acme_certificate:
    account_key_content: "{{ account_private_key }}"
    csr: /etc/pki/cert/csr/sample.com.csr
    dest: /etc/httpd/ssl/sample.com.crt
    modify_account: false
  register: sample_com_challenge

# Alternative first step:
- name: Create a challenge for sample.com using a account key from Hashi Vault.
  community.crypto.acme_certificate:
    account_key_content: >-
      {{ lookup('community.hashi_vault.hashi_vault', 'secret=secret/account_private_key:value') }}
    csr: /etc/pki/cert/csr/sample.com.csr
    fullchain_dest: /etc/httpd/ssl/sample.com-fullchain.crt
    modify_account: false
  register: sample_com_challenge

# Alternative first step:
- name: Create a challenge for sample.com using a account key file.
  community.crypto.acme_certificate:
    account_key_src: /etc/pki/cert/private/account.key
    csr_content: "{{ lookup('file', '/etc/pki/cert/csr/sample.com.csr') }}"
    dest: /etc/httpd/ssl/sample.com.crt
    fullchain_dest: /etc/httpd/ssl/sample.com-fullchain.crt
    modify_account: false
  register: sample_com_challenge

# perform the necessary steps to fulfill the challenge
# for example:
#
# - name: Copy http-01 challenge for sample.com
#   ansible.builtin.copy:
#     dest: /var/www/html/{{ sample_com_challenge['challenge_data']['sample.com']['http-01']['resource'] }}
#     content: "{{ sample_com_challenge['challenge_data']['sample.com']['http-01']['resource_value'] }}"
#   when: sample_com_challenge is changed and 'sample.com' in sample_com_challenge['challenge_data']
#
# Alternative way:
#
# - name: Copy http-01 challenges
#   ansible.builtin.copy:
#     dest: /var/www/{{ item.key }}/{{ item.value['http-01']['resource'] }}
#     content: "{{ item.value['http-01']['resource_value'] }}"
#   loop: "{{ sample_com_challenge.challenge_data | dict2items }}"
#   when: sample_com_challenge is changed

- name: Let the challenge be validated and retrieve the cert and intermediate certificate
  community.crypto.acme_certificate:
    account_key_src: /etc/pki/cert/private/account.key
    csr: /etc/pki/cert/csr/sample.com.csr
    dest: /etc/httpd/ssl/sample.com.crt
    fullchain_dest: /etc/httpd/ssl/sample.com-fullchain.crt
    chain_dest: /etc/httpd/ssl/sample.com-intermediate.crt
    data: "{{ sample_com_challenge }}"
    modify_account: false

---
### Example with DNS challenge against production ACME server ###

- name: Create a challenge for sample.com using a account key file.
  community.crypto.acme_certificate:
    account_key_src: /etc/pki/cert/private/account.key
    account_email: myself@sample.com
    src: /etc/pki/cert/csr/sample.com.csr
    cert: /etc/httpd/ssl/sample.com.crt
    challenge: dns-01
    acme_directory: https://acme-v01.api.letsencrypt.org/directory
    # Renew if the certificate is at least 30 days old
    remaining_days: 60
    modify_account: false
  register: sample_com_challenge

# perform the necessary steps to fulfill the challenge
# for example:
#
# - name: Create DNS record for sample.com dns-01 challenge
#   community.aws.route53:
#     zone: sample.com
#     record: "{{ sample_com_challenge.challenge_data['sample.com']['dns-01'].record }}"
#     type: TXT
#     ttl: 60
#     state: present
#     wait: true
#     # Note: route53 requires TXT entries to be enclosed in quotes
#     value: "{{ sample_com_challenge.challenge_data['sample.com']['dns-01'].resource_value | community.dns.quote_txt(always_quote=true) }}"
#   when: sample_com_challenge is changed and 'sample.com' in sample_com_challenge.challenge_data
#
# Alternative way:
#
# - name: Create DNS records for dns-01 challenges
#   community.aws.route53:
#     zone: sample.com
#     record: "{{ item.key }}"
#     type: TXT
#     ttl: 60
#     state: present
#     wait: true
#     # Note: item.value is a list of TXT entries, and route53
#     # requires every entry to be enclosed in quotes
#     value: "{{ item.value | map('community.dns.quote_txt', always_quote=true) | list }}"
#   loop: "{{ sample_com_challenge.challenge_data_dns | dict2items }}"
#   when: sample_com_challenge is changed

- name: Let the challenge be validated and retrieve the cert and intermediate certificate
  community.crypto.acme_certificate:
    account_key_src: /etc/pki/cert/private/account.key
    account_email: myself@sample.com
    src: /etc/pki/cert/csr/sample.com.csr
    cert: /etc/httpd/ssl/sample.com.crt
    fullchain: /etc/httpd/ssl/sample.com-fullchain.crt
    chain: /etc/httpd/ssl/sample.com-intermediate.crt
    challenge: dns-01
    acme_directory: https://acme-v01.api.letsencrypt.org/directory
    remaining_days: 60
    data: "{{ sample_com_challenge }}"
    modify_account: false
  when: sample_com_challenge is changed

# Alternative second step:
- name: Let the challenge be validated and retrieve the cert and intermediate certificate
  community.crypto.acme_certificate:
    account_key_src: /etc/pki/cert/private/account.key
    account_email: myself@sample.com
    src: /etc/pki/cert/csr/sample.com.csr
    cert: /etc/httpd/ssl/sample.com.crt
    fullchain: /etc/httpd/ssl/sample.com-fullchain.crt
    chain: /etc/httpd/ssl/sample.com-intermediate.crt
    challenge: tls-alpn-01
    remaining_days: 60
    data: "{{ sample_com_challenge }}"
    # We use Let's Encrypt's ACME v2 endpoint
    acme_directory: https://acme-v02.api.letsencrypt.org/directory
    # The following makes sure that if a chain with /CN=DST Root CA X3 in its issuer is provided
    # as an alternative, it will be selected. These are the roots cross-signed by IdenTrust.
    # As long as Let's Encrypt provides alternate chains with the cross-signed root(s) when
    # switching to their own ISRG Root X1 root, this will use the chain ending with a cross-signed
    # root. This chain is more compatible with older TLS clients.
    select_chain:
      - test_certificates: last
        issuer:
          CN: DST Root CA X3
          O: Digital Signature Trust Co.
    modify_account: false
  when: sample_com_challenge is changed
"""

RETURN = r"""
cert_days:
  description: The number of days the certificate remains valid.
  returned: success
  type: int
challenge_data:
  description:
    - Per identifier / challenge type challenge data.
    - Since Ansible 2.8.5, only challenges which are not yet valid are returned.
  returned: changed
  type: dict
  contains:
    identifier:
      description:
        - For every identifier, provides a dictionary of challenge types mapping to challenge data.
        - The keys in this dictionary are the identifiers. C(identifier) is a placeholder used in the documentation.
        - Note that the keys are not valid Jinja2 identifiers.
      returned: changed
      type: dict
      contains:
        challenge-type:
          description:
            - Data for every challenge type.
            - The keys in this dictionary are the challenge types. C(challenge-type) is a placeholder used in the documentation.
              Possible keys are V(http-01), V(dns-01), V(dns-account-01), V(dns-persist-01), and V(tls-alpn-01).
            - Note that the keys are not valid Jinja2 identifiers.
            - B(Note) that this return value for V(dns-account-01) and V(dns-persist-01) challenges is B(experimental)
              and can change in a breaking way also in minor releases, until the acme-dns-account-label respectively
              the acme-dns-persist draft is in a more stable state.
          returned: changed
          type: dict
          contains:
            resource:
              description: The challenge resource that must be created for validation.
              returned: changed and challenge is not V(dns-persist-01)
              type: str
              sample: .well-known/acme-challenge/evaGxfADs6pSRb2LAv9IZf17Dt3juxGJ-PCt92wr-oA
            resource_original:
              description:
                - The original challenge resource including type identifier for V(tls-alpn-01) challenges.
              returned: changed and O(challenge) is V(tls-alpn-01)
              type: str
              sample: DNS:example.com
            resource_value:
              description:
                - The value the resource has to produce for the validation.
                - For V(http-01), V(dns-01), and V(dns-account-01) challenges, the value can be used as-is.
                - For V(tls-alpn-01) challenges, note that this return value contains a Base64 encoded version of the correct
                  binary blob which has to be put into the acmeValidation x509 extension; see U(https://www.rfc-editor.org/rfc/rfc8737.html#section-3)
                  for details. To do this, you might need the P(ansible.builtin.b64decode#filter) Jinja filter to extract
                  the binary blob from this return value.
              returned: changed and challenge is not V(dns-persist-01)
              type: str
              sample: IlirfxKKXA...17Dt3juxGJ-PCt92wr-oA
            record:
              description: The full DNS record's name for the challenge.
              returned: changed and challenge is V(dns-01) or V(dns-account-01)
              type: str
              sample: _acme-challenge.example.com
            account_uri:
              description:
                - The account URI that must be mentioned in the DNS TXT record.
                - B(Note) that the type of this return value might change in a breaking way also in minor releases,
                  until the acme-dns-persist draft is in a more stable state.
                  This challenge field is currently under discussion in the ACME WG.
              returned: changed and challenge is V(dns-persist-01)
              type: str
              sample: https://ca.example/acct/123
            issuer_domain_names:
              description:
                - One of the issuer domain names must be mentioned in the DNS TXT record.
              returned: changed and challenge is V(dns-persist-01)
              type: list
              elements: str
              sample:
                - letsencrypt.org
challenge_data_dns:
  description:
    - List of TXT values per DNS record, in case challenge is V(dns-01) or V(dns-account-01).
    - Since Ansible 2.8.5, only challenges which are not yet valid are returned.
  returned: changed
  type: dict
authorizations:
  description:
    - ACME authorization data.
    - Maps an identifier to ACME authorization objects. See U(https://tools.ietf.org/html/rfc8555#section-7.1.4).
  returned: changed
  type: dict
  sample:
    example.com:
      identifier:
        type: dns
        value: example.com
      status: valid
      expires: '2022-08-04T01:02:03.45Z'
      challenges:
        - url: https://example.org/acme/challenge/12345
          type: http-01
          status: valid
          token: A5b1C3d2E9f8G7h6
          validated: '2022-08-01T01:01:02.34Z'
      wildcard: false
order_uri:
  description: ACME order URI.
  returned: changed
  type: str
finalization_uri:
  description: ACME finalization URI.
  returned: changed
  type: str
account_uri:
  description: ACME account URI.
  returned: changed
  type: str
all_chains:
  description:
    - When O(retrieve_all_alternates) is set to V(true), the module will query the ACME server for alternate chains. This
      return value will contain a list of all chains returned, the first entry being the main chain returned by the server.
    - See L(Section 7.4.2 of RFC8555,https://tools.ietf.org/html/rfc8555#section-7.4.2) for details.
  returned: when certificate was retrieved and O(retrieve_all_alternates) is set to V(true)
  type: list
  elements: dict
  contains:
    cert:
      description:
        - The leaf certificate itself, in PEM format.
      type: str
      returned: always
    chain:
      description:
        - The certificate chain, excluding the root, as concatenated PEM certificates.
      type: str
      returned: always
    full_chain:
      description:
        - The certificate chain, excluding the root, but including the leaf certificate, as concatenated PEM certificates.
      type: str
      returned: always
"""

import os
import typing as t

from ansible_collections.community.crypto.plugins.module_utils._acme.account import (
    ACMEAccount,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.acme import (
    ACMEClient,
    create_backend,
    create_default_argspec,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.certificates import (
    CertificateChain,
    Criterium,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.challenges import (
    combine_identifier,
    normalize_combined_identifier,
    wait_for_validation,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.errors import (
    ModuleFailException,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.io import (
    write_file,
)
from ansible_collections.community.crypto.plugins.module_utils._acme.orders import Order
from ansible_collections.community.crypto.plugins.module_utils._acme.utils import (
    compute_cert_id,
    pem_to_der,
)

if t.TYPE_CHECKING:
    from ansible.module_utils.basic import AnsibleModule  # pragma: no cover

    from ansible_collections.community.crypto.plugins.module_utils._acme.backends import (  # pragma: no cover
        CertificateInformation,
        CryptoBackend,
    )
    from ansible_collections.community.crypto.plugins.module_utils._acme.challenges import (  # pragma: no cover
        Authorization,
    )


NO_CHALLENGE = "no challenge"


class ACMECertificateClient:
    """
    ACME client class. Uses an ACME account object and a CSR to
    start and validate ACME challenges and download the respective
    certificates.
    """

    def __init__(self, module: AnsibleModule, backend: CryptoBackend):
        self.module = module
        self.version = module.params["acme_version"]
        self.challenge = module.params["challenge"]
        # We use None instead of a magic string for 'no challenge'
        if self.challenge == NO_CHALLENGE:
            self.challenge = None
        self.csr = module.params["csr"]
        self.csr_content = module.params["csr_content"]
        self.dest = module.params.get("dest")
        self.fullchain_dest = module.params.get("fullchain_dest")
        self.chain_dest = module.params.get("chain_dest")
        self.client = ACMEClient(module=module, backend=backend)
        self.account = ACMEAccount(client=self.client)
        self.directory = self.client.directory
        self.data = module.params["data"]
        self.authorizations: dict[str, Authorization] | None = None
        self.cert_days = -1
        self.order: Order | None = None
        self.order_uri = self.data.get("order_uri") if self.data else None
        self.all_chains: list[dict[str, t.Any]] | None = None
        self.select_chain_matcher = []
        self.include_renewal_cert_id = module.params["include_renewal_cert_id"]
        self.profile = module.params["profile"]
        self.order_creation_error_strategy = module.params[
            "order_creation_error_strategy"
        ]
        self.order_creation_max_retries = module.params["order_creation_max_retries"]

        if self.module.params["select_chain"]:
            for criterium_idx, criterium in enumerate(
                self.module.params["select_chain"]
            ):
                try:
                    self.select_chain_matcher.append(
                        self.client.backend.create_chain_matcher(
                            criterium=Criterium(
                                criterium=criterium, index=criterium_idx
                            )
                        )
                    )
                except ValueError as exc:
                    self.module.warn(
                        f"Error while parsing criterium: {exc}. Ignoring criterium."
                    )

        if self.profile is not None:
            meta_profiles = (self.directory.get("meta") or {}).get("profiles") or {}
            if not meta_profiles:
                raise ModuleFailException(msg="The ACME CA does not support profiles.")
            if self.profile not in meta_profiles:
                raise ModuleFailException(
                    msg=f"The ACME CA does not support selected profile {self.profile!r}."
                )

        # Make sure account exists
        modify_account = module.params["modify_account"]
        if modify_account is None:
            module.deprecate(
                "The default 'true' for modify_account has been deprecated."
                " The default will change to 'false' in community.crypto 4.0.0."
                " We suggest to explicitly set this option to a value to avoid"
                " this warning. We also recommend to not set it to 'true',"
                " but to use the community.crypto.acme_account module instead.",
                version="4.0.0",
                collection_name="community.crypto",
            )

            modify_account = True
        contact = []
        if module.params["account_email"]:
            contact.append("mailto:" + module.params["account_email"])
        created, account_data = self.account.setup_account(
            contact=contact,
            terms_agreed=module.params.get("terms_agreed"),
            allow_creation=modify_account,
        )
        if account_data is None:
            raise ModuleFailException(msg="Account does not exist or is deactivated.")
        updated = False
        if not created and account_data and modify_account:
            updated, account_data = self.account.update_account(
                account_data=account_data, contact=contact
            )
        self.changed = created or updated

        if self.csr is not None and not os.path.exists(self.csr):
            raise ModuleFailException(f"CSR {self.csr} not found")

        # Extract list of identifiers from CSR
        self.identifiers = self.client.backend.get_ordered_csr_identifiers(
            csr_filename=self.csr, csr_content=self.csr_content
        )

    def is_first_step(self) -> bool:
        """
        Return True if this is the first execution of this module, i.e. if a
        sufficient data object from a first run has not been provided.
        """
        if self.data is None:
            return True
        # We are in the second stage if data.order_uri is given (which has been
        # stored in self.order_uri by the constructor).
        return self.order_uri is None

    def _get_cert_info_or_none(self) -> CertificateInformation | None:
        if self.module.params.get("dest"):
            filename = self.module.params["dest"]
        else:
            filename = self.module.params["fullchain_dest"]
        if not os.path.exists(filename):
            return None
        return self.client.backend.get_cert_information(cert_filename=filename)

    def start_challenges(self) -> None:
        """
        Create new authorizations for all identifiers of the CSR,
        respectively start a new order for ACME v2.
        """
        self.authorizations = {}
        replaces_cert_id = None
        if self.include_renewal_cert_id == "always" or (
            self.include_renewal_cert_id == "when_ari_supported"
            and self.client.directory.has_renewal_info_endpoint()
        ):
            cert_info = self._get_cert_info_or_none()
            if cert_info is not None:
                replaces_cert_id = compute_cert_id(
                    backend=self.client.backend,
                    cert_info=cert_info,
                    none_if_required_information_is_missing=True,
                )
        self.order = Order.create_with_error_handling(
            client=self.client,
            identifiers=self.identifiers,
            error_strategy=self.order_creation_error_strategy,
            error_max_retries=self.order_creation_max_retries,
            replaces_cert_id=replaces_cert_id,
            profile=self.profile,
            message_callback=self.module.warn,
        )
        self.order_uri = self.order.url
        self.order.load_authorizations(client=self.client)
        self.authorizations.update(self.order.authorizations)
        self.changed = True

    def get_challenges_data(
        self, first_step: bool
    ) -> tuple[dict[str, t.Any], dict[str, list[str]]]:
        """
        Get challenge details for the chosen challenge type.
        Return a tuple of generic challenge details, and specialized DNS challenge details.
        """
        assert self.authorizations is not None
        data: dict[str, t.Any] = {}
        data_dns: dict[str, list[str]] = {}
        for type_identifier, authz in self.authorizations.items():
            # Skip valid authentications: their challenges are already valid
            # and do not need to be returned
            if authz.status == "valid":
                continue
            # We drop the type from the key to preserve backwards compatibility
            challenges = authz.get_challenge_data(client=self.client)
            assert authz.identifier is not None
            data[authz.identifier] = challenges
            if (
                first_step
                and self.challenge is not None
                and self.challenge not in data[authz.identifier]
            ):
                raise ModuleFailException(
                    f"Found no challenge of type '{self.challenge}' for identifier {type_identifier}!"
                )
            if (
                self.challenge in ("dns-01", "dns-account-01")
                and self.challenge in challenges
            ):
                values = data_dns.get(challenges[self.challenge]["record"])
                if values is None:
                    values = []
                    data_dns[challenges[self.challenge]["record"]] = values
                values.append(challenges[self.challenge]["resource_value"])
        return data, data_dns

    def finish_challenges(self) -> None:
        """
        Verify challenges for all identifiers of the CSR.
        """
        self.authorizations = {}

        # Step 1: obtain challenge information
        # For ACME v2, we obtain the order object by fetching the
        # order URI, and extract the information from there.
        assert self.order_uri is not None
        self.order = Order.from_url(client=self.client, url=self.order_uri)
        self.order.load_authorizations(client=self.client)
        self.authorizations.update(self.order.authorizations)

        # Step 2: validate pending challenges
        authzs_to_wait_for = []
        for authz in self.authorizations.values():
            if authz.status == "pending":
                if self.challenge is not None:
                    authz.call_validate(
                        client=self.client, challenge_type=self.challenge, wait=False
                    )
                    authzs_to_wait_for.append(authz)
                # If there is no challenge, we must check whether the authz is valid
                elif authz.status != "valid":
                    authz.raise_error(
                        error_msg='Status is not "valid", even though no challenge should be necessary',
                        module=self.client.module,
                    )
                self.changed = True

        # Step 3: wait for authzs to validate
        wait_for_validation(authzs=authzs_to_wait_for, client=self.client)

    def download_alternate_chains(
        self, cert: CertificateChain
    ) -> list[CertificateChain]:
        alternate_chains = []
        for alternate in cert.alternates:
            try:
                alt_cert = CertificateChain.download(client=self.client, url=alternate)
            except ModuleFailException as e:
                self.module.warn(
                    f"Error while downloading alternative certificate {alternate}: {e}"
                )
                continue
            alternate_chains.append(alt_cert)
        return alternate_chains

    def find_matching_chain(
        self, chains: t.Iterable[CertificateChain]
    ) -> CertificateChain | None:
        for criterium_idx, matcher in enumerate(self.select_chain_matcher):
            for chain in chains:
                if matcher.match(certificate=chain):
                    self.module.debug(
                        f"Found matching chain for criterium {criterium_idx}"
                    )
                    return chain
        return None

    def get_certificate(self) -> None:
        """
        Request a new certificate and write it to the destination file.
        First verifies whether all authorizations are valid; if not, aborts
        with an error.
        """
        assert self.authorizations is not None
        for identifier_type, identifier in self.identifiers:
            authz = self.authorizations.get(
                normalize_combined_identifier(
                    combine_identifier(
                        identifier_type=identifier_type, identifier=identifier
                    )
                )
            )
            if authz is None:
                raise ModuleFailException(
                    f'Found no authorization information for "{combine_identifier(identifier_type=identifier_type, identifier=identifier)}"!'
                )
            if authz.status != "valid":
                authz.raise_error(
                    error_msg=f'Status is "{authz.status}" and not "valid"',
                    module=self.module,
                )

        assert self.order is not None
        self.order.finalize(
            client=self.client,
            csr_der=pem_to_der(pem_filename=self.csr, pem_content=self.csr_content),
        )
        assert self.order.certificate_uri is not None
        cert = CertificateChain.download(
            client=self.client, url=self.order.certificate_uri
        )
        if self.module.params["retrieve_all_alternates"] or self.select_chain_matcher:
            # Retrieve alternate chains
            alternate_chains = self.download_alternate_chains(cert)

            # Prepare return value for all alternate chains
            if self.module.params["retrieve_all_alternates"]:
                self.all_chains = [cert.to_json()]
                for alt_chain in alternate_chains:
                    self.all_chains.append(alt_chain.to_json())

            # Try to select alternate chain depending on criteria
            if self.select_chain_matcher:
                matching_chain = self.find_matching_chain([cert] + alternate_chains)
                if matching_chain:
                    cert = matching_chain
                else:
                    self.module.debug("Found no matching alternative chain")

        if cert.cert is not None:
            pem_cert = cert.cert
            chain = cert.chain

            if self.dest and write_file(
                module=self.module, dest=self.dest, content=pem_cert.encode("utf8")
            ):
                self.cert_days = self.client.backend.get_cert_days(
                    cert_filename=self.dest
                )
                self.changed = True

            if self.fullchain_dest and write_file(
                module=self.module,
                dest=self.fullchain_dest,
                content=(pem_cert + "\n".join(chain)).encode("utf8"),
            ):
                self.cert_days = self.client.backend.get_cert_days(
                    cert_filename=self.fullchain_dest
                )
                self.changed = True

            if self.chain_dest and write_file(
                module=self.module,
                dest=self.chain_dest,
                content=("\n".join(chain)).encode("utf8"),
            ):
                self.changed = True

    def deactivate_authzs(self) -> None:
        """
        Deactivates all valid authz's. Does not raise exceptions.
        https://community.letsencrypt.org/t/authorization-deactivation/19860/2
        https://tools.ietf.org/html/rfc8555#section-7.5.2
        """
        assert self.authorizations is not None
        for authz in self.authorizations.values():
            try:
                authz.deactivate(client=self.client)
            except Exception:
                # ignore errors
                pass
            if not authz.is_in_final_state(allow_valid=False):
                self.module.warn(
                    warning=f"Could not deactivate authz object {authz.url}."
                )


def main() -> t.NoReturn:
    argument_spec = create_default_argspec(with_certificate=True)
    argument_spec.argument_spec["csr"]["aliases"] = ["src"]
    argument_spec.update_argspec(
        modify_account={"type": "bool"},
        account_email={"type": "str"},
        agreement={
            "type": "str",
            "removed_in_version": "4.0.0",
            "removed_from_collection": "community.crypto",
        },
        terms_agreed={"type": "bool", "default": False},
        challenge={
            "type": "str",
            "default": "http-01",
            "choices": [
                "http-01",
                "dns-01",
                "dns-account-01",
                "dns-persist-01",
                "tls-alpn-01",
                NO_CHALLENGE,
            ],
        },
        data={"type": "dict"},
        dest={"type": "path", "aliases": ["cert"]},
        fullchain_dest={"type": "path", "aliases": ["fullchain"]},
        chain_dest={"type": "path", "aliases": ["chain"]},
        remaining_days={"type": "int", "default": 10},
        deactivate_authzs={"type": "bool", "default": False},
        force={"type": "bool", "default": False},
        retrieve_all_alternates={"type": "bool", "default": False},
        select_chain={
            "type": "list",
            "elements": "dict",
            "options": {
                "test_certificates": {
                    "type": "str",
                    "default": "all",
                    "choices": ["first", "last", "all"],
                },
                "issuer": {"type": "dict"},
                "subject": {"type": "dict"},
                "subject_key_identifier": {"type": "str"},
                "authority_key_identifier": {"type": "str"},
            },
        },
        include_renewal_cert_id={
            "type": "str",
            "choices": ["never", "when_ari_supported", "always"],
            "default": "never",
        },
        profile={"type": "str"},
        order_creation_error_strategy={
            "type": "str",
            "default": "auto",
            "choices": ["auto", "always", "fail", "retry_without_replaces_cert_id"],
        },
        order_creation_max_retries={"type": "int", "default": 3},
    )
    argument_spec.update(
        required_one_of=[
            ["dest", "fullchain_dest"],
        ],
    )
    module = argument_spec.create_ansible_module(supports_check_mode=True)
    backend = create_backend(module, needs_acme_v2=False)

    try:
        if module.params.get("dest"):
            cert_days = backend.get_cert_days(cert_filename=module.params["dest"])
        else:
            cert_days = backend.get_cert_days(
                cert_filename=module.params["fullchain_dest"]
            )

        if module.params["force"] or cert_days < module.params["remaining_days"]:
            # If checkmode is active, base the changed state solely on the status
            # of the certificate file as all other actions (accessing an account, checking
            # the authorization status...) would lead to potential changes of the current
            # state
            if module.check_mode:
                module.exit_json(
                    changed=True,
                    authorizations={},
                    challenge_data={},
                    cert_days=cert_days,
                )
            else:
                client = ACMECertificateClient(module=module, backend=backend)
                client.cert_days = cert_days
                other: dict[str, t.Any] = {}
                is_first_step = client.is_first_step()
                if is_first_step:
                    # First run: start challenges / start new order
                    client.start_challenges()
                else:
                    # Second run: finish challenges, and get certificate
                    try:
                        client.finish_challenges()
                        client.get_certificate()
                        if client.all_chains is not None:
                            other["all_chains"] = client.all_chains
                    finally:
                        if module.params["deactivate_authzs"]:
                            client.deactivate_authzs()
                data, data_dns = client.get_challenges_data(first_step=is_first_step)
                auths = {}
                assert client.authorizations is not None
                for v in client.authorizations.values():
                    # Remove "type:" from key
                    auths[v.identifier] = v.to_json()
                module.exit_json(
                    changed=client.changed,
                    authorizations=auths,
                    finalize_uri=client.order.finalize_uri if client.order else None,
                    order_uri=client.order_uri,
                    account_uri=client.client.account_uri,
                    challenge_data=data,
                    challenge_data_dns=data_dns,
                    cert_days=client.cert_days,
                    **other,
                )
        else:
            module.exit_json(changed=False, cert_days=cert_days)
    except ModuleFailException as e:
        e.do_fail(module=module)


if __name__ == "__main__":
    main()



#!/usr/bin/python

# Copyright (c) Ansible project
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: keycloak_client_rolescope

short_description: Allows administration of Keycloak client roles scope to restrict the usage of certain roles to a other
  specific client applications

version_added: 8.6.0

description:
  - This module allows you to add or remove Keycloak roles from clients scope using the Keycloak REST API. It requires access
    to the REST API using OpenID Connect; the user connecting and the client being used must have the requisite access rights.
    In a default Keycloak installation, admin-cli and an admin user would work, as would a separate client definition with
    the scope tailored to your needs and a user having the expected roles.
  - Client O(client_id) must have O(community.general.keycloak_client#module:full_scope_allowed) set to V(false).
  - Attributes are multi-valued in the Keycloak API. All attributes are lists of individual values and are returned that way
    by this module. You may pass single values for attributes when calling the module, and this is translated into a list
    suitable for the API.
attributes:
  check_mode:
    support: full
  diff_mode:
    support: full
  action_group:
    version_added: 10.2.0

options:
  state:
    description:
      - State of the role mapping.
      - On V(present), all roles in O(role_names) are mapped if not exist yet.
      - On V(absent), all roles mapping in O(role_names) are removed if it exists.
    default: 'present'
    type: str
    choices:
      - present
      - absent

  realm:
    type: str
    description:
      - The Keycloak realm under which clients resides.
    default: 'master'

  client_id:
    type: str
    required: true
    description:
      - Roles provided in O(role_names) while be added to this client scope.
  client_scope_id:
    type: str
    description:
      - If the O(role_names) are client role, the client ID under which it resides.
      - If this parameter is absent, the roles are considered a realm role.
  role_names:
    required: true
    type: list
    elements: str
    description:
      - Names of roles to manipulate.
      - If O(client_scope_id) is present, all roles must be under this client.
      - If O(client_scope_id) is absent, all roles must be under the realm.
extends_documentation_fragment:
  - community.general._keycloak
  - community.general._keycloak.actiongroup_keycloak
  - community.general._attributes

author:
  - Andre Desrosiers (@desand01)
"""

EXAMPLES = r"""
- name: Add roles to public client scope
  community.general.keycloak_client_rolescope:
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
    realm: MyCustomRealm
    client_id: frontend-client-public
    client_scope_id: backend-client-private
    role_names:
      - backend-role-admin
      - backend-role-user

- name: Remove roles from public client scope
  community.general.keycloak_client_rolescope:
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
    realm: MyCustomRealm
    client_id: frontend-client-public
    client_scope_id: backend-client-private
    role_names:
      - backend-role-admin
    state: absent

- name: Add realm roles to public client scope
  community.general.keycloak_client_rolescope:
    auth_keycloak_url: https://auth.example.com/auth
    auth_realm: master
    auth_username: USERNAME
    auth_password: PASSWORD
    realm: MyCustomRealm
    client_id: frontend-client-public
    role_names:
      - realm-role-admin
      - realm-role-user
"""

RETURN = r"""
msg:
  description: Message as to what action was taken.
  returned: always
  type: str
  sample: "Client role scope for frontend-client-public has been updated"

end_state:
  description: Representation of role role scope after module execution.
  returned: on success
  type: list
  elements: dict
  sample:
    [
      {
        "clientRole": false,
        "composite": false,
        "containerId": "MyCustomRealm",
        "id": "47293104-59a6-46f0-b460-2e9e3c9c424c",
        "name": "backend-role-admin"
      },
      {
        "clientRole": false,
        "composite": false,
        "containerId": "MyCustomRealm",
        "id": "39c62a6d-542c-4715-92d2-41021eb33967",
        "name": "backend-role-user"
      }
    ]
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.general.plugins.module_utils._keycloak import (
    KeycloakAPI,
    KeycloakError,
    get_token,
    keycloak_argument_spec,
)


def main():
    """
    Module execution

    :return:
    """
    argument_spec = keycloak_argument_spec()

    meta_args = dict(
        client_id=dict(type="str", required=True),
        client_scope_id=dict(type="str"),
        realm=dict(type="str", default="master"),
        role_names=dict(type="list", elements="str", required=True),
        state=dict(type="str", default="present", choices=["present", "absent"]),
    )

    argument_spec.update(meta_args)

    module = AnsibleModule(argument_spec=argument_spec, supports_check_mode=True)

    result = dict(changed=False, msg="", diff={}, end_state={})

    # Obtain access token, initialize API
    try:
        connection_header = get_token(module.params)
    except KeycloakError as e:
        module.fail_json(msg=str(e))

    kc = KeycloakAPI(module, connection_header)

    realm = module.params.get("realm")
    clientid = module.params.get("client_id")
    client_scope_id = module.params.get("client_scope_id")
    role_names = module.params.get("role_names")
    state = module.params.get("state")

    objRealm = kc.get_realm_by_id(realm)
    if not objRealm:
        module.fail_json(msg=f"Failed to retrive realm '{realm}'")

    objClient = kc.get_client_by_clientid(clientid, realm)
    if not objClient:
        module.fail_json(msg=f"Failed to retrive client '{realm}.{clientid}'")
    if objClient["fullScopeAllowed"] and state == "present":
        module.fail_json(msg=f"FullScopeAllowed is active for Client '{realm}.{clientid}'")

    if client_scope_id:
        objClientScope = kc.get_client_by_clientid(client_scope_id, realm)
        if not objClientScope:
            module.fail_json(msg=f"Failed to retrive client '{realm}.{client_scope_id}'")
        before_role_mapping = kc.get_client_role_scope_from_client(objClient["id"], objClientScope["id"], realm)
    else:
        before_role_mapping = kc.get_client_role_scope_from_realm(objClient["id"], realm)

    if client_scope_id:
        # retrive all role from client_scope
        client_scope_roles_by_name = kc.get_client_roles_by_id(objClientScope["id"], realm)
    else:
        # retrive all role from realm
        client_scope_roles_by_name = kc.get_realm_roles(realm)

    # convert to indexed Dict by name
    client_scope_roles_by_name = {role["name"]: role for role in client_scope_roles_by_name}
    role_mapping_by_name = {role["name"]: role for role in before_role_mapping}
    role_mapping_to_manipulate = []

    if state == "present":
        # update desired
        for role_name in role_names:
            if role_name not in client_scope_roles_by_name:
                if client_scope_id:
                    module.fail_json(msg=f"Failed to retrive role '{realm}.{client_scope_id}.{role_name}'")
                else:
                    module.fail_json(msg=f"Failed to retrive role '{realm}.{role_name}'")
            if role_name not in role_mapping_by_name:
                role_mapping_to_manipulate.append(client_scope_roles_by_name[role_name])
                role_mapping_by_name[role_name] = client_scope_roles_by_name[role_name]
    else:
        # remove role if present
        for role_name in role_names:
            if role_name in role_mapping_by_name:
                role_mapping_to_manipulate.append(role_mapping_by_name[role_name])
                del role_mapping_by_name[role_name]

    before_role_mapping = sorted(before_role_mapping, key=lambda d: d["name"])
    desired_role_mapping = sorted(role_mapping_by_name.values(), key=lambda d: d["name"])

    result["changed"] = len(role_mapping_to_manipulate) > 0

    if result["changed"]:
        result["diff"] = dict(before=before_role_mapping, after=desired_role_mapping)

    if not result["changed"]:
        # no changes
        result["end_state"] = before_role_mapping
        result["msg"] = f"No changes required for client role scope {clientid}."
    elif state == "present":
        # doing update
        if module.check_mode:
            result["end_state"] = desired_role_mapping
        elif client_scope_id:
            result["end_state"] = kc.update_client_role_scope_from_client(
                role_mapping_to_manipulate, objClient["id"], objClientScope["id"], realm
            )
        else:
            result["end_state"] = kc.update_client_role_scope_from_realm(
                role_mapping_to_manipulate, objClient["id"], realm
            )
        result["msg"] = f"Client role scope for {clientid} has been updated"
    else:
        # doing delete
        if module.check_mode:
            result["end_state"] = desired_role_mapping
        elif client_scope_id:
            result["end_state"] = kc.delete_client_role_scope_from_client(
                role_mapping_to_manipulate, objClient["id"], objClientScope["id"], realm
            )
        else:
            result["end_state"] = kc.delete_client_role_scope_from_realm(
                role_mapping_to_manipulate, objClient["id"], realm
            )
        result["msg"] = f"Client role scope for {clientid} has been deleted"
    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# Copyright (c) 2021, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import annotations

DOCUMENTATION = r"""
module: openssl_publickey_info
short_description: Provide information for OpenSSL public keys
description:
  - This module allows one to query information on OpenSSL public keys.
  - It uses the cryptography python library to interact with OpenSSL.
version_added: 1.7.0
author:
  - Felix Fontein (@felixfontein)
extends_documentation_fragment:
  - community.crypto._attributes
  - community.crypto._attributes.info_module
  - community.crypto._attributes.idempotent_not_modify_state
  - community.crypto._cryptography_dep.minimum
options:
  path:
    description:
      - Remote absolute path where the public key file is loaded from.
    type: path
  content:
    description:
      - Content of the public key file.
      - Either O(path) or O(content) must be specified, but not both.
    type: str

  select_crypto_backend:
    description:
      - Determines which crypto backend to use.
      - The default choice is V(auto), which tries to use C(cryptography) if available.
      - If set to V(cryptography), will try to use the L(cryptography,https://cryptography.io/) library.
      - Note that with community.crypto 3.0.0, all values behave the same.
        This option will be deprecated in a later version.
        We recommend to not set it explicitly.
    type: str
    default: auto
    choices: [auto, cryptography]

seealso:
  - module: community.crypto.openssl_publickey
  - module: community.crypto.openssl_privatekey_info
  - plugin: community.crypto.openssl_publickey_info
    plugin_type: filter
    description: A filter variant of this module.
"""

EXAMPLES = r"""
---
- name: Generate an OpenSSL private key with the default values (4096 bits, RSA)
  community.crypto.openssl_privatekey:
    path: /etc/ssl/private/ansible.com.pem

- name: Create public key from private key
  community.crypto.openssl_publickey:
    privatekey_path: /etc/ssl/private/ansible.com.pem
    path: /etc/ssl/ansible.com.pub

- name: Get information on public key
  community.crypto.openssl_publickey_info:
    path: /etc/ssl/ansible.com.pub
  register: result

- name: Dump information
  ansible.builtin.debug:
    var: result
"""

RETURN = r"""
fingerprints:
  description:
    - Fingerprints of public key.
    - For every hash algorithm available, the fingerprint is computed.
  returned: success
  type: dict
  sample: "{'sha256': 'd4:b3:aa:6d:c8:04:ce:4e:ba:f6:29:4d:92:a3:94:b0:c2:ff:bd:bf:33:63:11:43:34:0f:51:b0:95:09:2f:63', 'sha512':
    'f7:07:4a:f0:b0:f0:e6:8b:95:5f:f9:e6:61:0a:32:68:f1..."
type:
  description:
    - The key's type.
    - One of V(RSA), V(DSA), V(ECC), V(Ed25519), V(X25519), V(Ed448), or V(X448).
    - Will start with V(unknown) if the key type cannot be determined.
  returned: success
  type: str
  sample: RSA
public_data:
  description:
    - Public key data. Depends on key type.
  returned: success
  type: dict
  contains:
    size:
      description:
        - Bit size of modulus (RSA) or prime number (DSA).
      type: int
      returned: When RV(type=RSA) or RV(type=DSA)
    modulus:
      description:
        - The RSA key's modulus.
      type: int
      returned: When RV(type=RSA)
    exponent:
      description:
        - The RSA key's public exponent.
      type: int
      returned: When RV(type=RSA)
    p:
      description:
        - The C(p) value for DSA.
        - This is the prime modulus upon which arithmetic takes place.
      type: int
      returned: When RV(type=DSA)
    q:
      description:
        - The C(q) value for DSA.
        - This is a prime that divides C(p - 1), and at the same time the order of the subgroup of the multiplicative group
          of the prime field used.
      type: int
      returned: When RV(type=DSA)
    g:
      description:
        - The C(g) value for DSA.
        - This is the element spanning the subgroup of the multiplicative group of the prime field used.
      type: int
      returned: When RV(type=DSA)
    curve:
      description:
        - The curve's name for ECC.
      type: str
      returned: When RV(type=ECC)
    exponent_size:
      description:
        - The maximum number of bits of a private key. This is basically the bit size of the subgroup used.
      type: int
      returned: When RV(type=ECC)
    x:
      description:
        - The C(x) coordinate for the public point on the elliptic curve.
      type: int
      returned: When RV(type=ECC)
    y:
      description:
        - For RV(type=ECC), this is the C(y) coordinate for the public point on the elliptic curve.
        - For RV(type=DSA), this is the publicly known group element whose discrete logarithm w.r.t. C(g) is the private key.
      type: int
      returned: When RV(type=DSA) or RV(type=ECC)
"""

import typing as t

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.community.crypto.plugins.module_utils._crypto.basic import (
    OpenSSLObjectError,
)
from ansible_collections.community.crypto.plugins.module_utils._crypto.module_backends.publickey_info import (
    PublicKeyParseError,
    select_backend,
)


def main() -> t.NoReturn:
    module = AnsibleModule(
        argument_spec={
            "path": {"type": "path"},
            "content": {"type": "str", "no_log": True},
            "select_crypto_backend": {
                "type": "str",
                "default": "auto",
                "choices": ["auto", "cryptography"],
            },
        },
        required_one_of=(["path", "content"],),
        mutually_exclusive=(["path", "content"],),
        supports_check_mode=True,
    )

    result = {
        "can_load_key": False,
        "can_parse_key": False,
        "key_is_consistent": None,
    }

    if module.params["content"] is not None:
        data = module.params["content"].encode("utf-8")
    else:
        try:
            with open(module.params["path"], "rb") as f:
                data = f.read()
        except (IOError, OSError) as e:
            module.fail_json(
                msg=f"Error while reading public key file from disk: {e}",
                **result,  # type: ignore
            )

    module_backend = select_backend(module=module, content=data)

    try:
        result.update(module_backend.get_info())
        module.exit_json(**result)
    except PublicKeyParseError as exc:
        result.update(exc.result)
        module.fail_json(msg=exc.error_message, **result)  # type: ignore
    except OpenSSLObjectError as exc:
        module.fail_json(msg=str(exc))


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2024, Sabari Jaganathan (@sajagana) <sajagana@cisco.com>

# GNU General Public License v3.0+ (see LICENSE or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function

__metaclass__ = type

ANSIBLE_METADATA = {"metadata_version": "1.1", "status": ["preview"], "supported_by": "community"}

DOCUMENTATION = r"""
---
module: mso_schema_template_anp_epg_intra_epg_contract
short_description: Manage Intra-EPG Contract on Cisco Nexus Dashboard Orchestrator (NDO).
description:
- Manage Intra-EPG Contract on Cisco Nexus Dashboard Orchestrator (NDO).
author:
- Sabari Jaganathan (@sajagana)
options:
  schema:
    description:
    - The name of the Schema.
    type: str
    required: true
  template:
    description:
    - The name of the Template.
    type: str
    required: true
  anp:
    description:
    - The name of the Application Profile.
    type: str
    aliases: [ application_profile ]
    required: true
  epg:
    description:
    - The name of the Endpoint Group.
    type: str
    aliases: [ endpoint_group ]
    required: true
  contract:
    description:
    - A contract associated to this Intra-EPG Contract.
    type: dict
    suboptions:
      name:
        description:
        - The name of the Contract to associate with.
        required: true
        type: str
        aliases: [ contract_name ]
      schema:
        description:
        - The schema that defines the referenced Contract.
        - If this parameter is unspecified, it defaults to the current O(schema).
        type: str
        aliases: [ contract_schema ]
      template:
        description:
        - The template that defines the referenced Contract.
        - If this parameter is unspecified, it defaults to the current O(template).
        type: str
        aliases: [ contract_template ]
  state:
    description:
    - Use C(absent) for removing.
    - Use C(query) for listing an object or multiple objects.
    - Use C(present) for creating or updating.
    type: str
    choices: [ absent, query, present ]
    default: query
notes:
- The O(schema) must exist before using this module in your playbook.
  Use M(cisco.mso.mso_schema) to create the Schema.
- The O(template) must exist before using this module in your playbook.
  Use M(cisco.mso.mso_schema_template) to create the Template.
- The O(anp) must exist before using this module in your playbook.
  Use M(cisco.mso.mso_schema_template_anp) to create the Application Profile.
- The O(epg) must exist before using this module in your playbook.
  Use M(cisco.mso.mso_schema_template_anp_epg) to create the Endpoint Group.
- The O(contract) must exist before using this module in your playbook.
  Use M(cisco.mso.mso_schema_template_contract_filter) to create the Contract with Filter.
seealso:
- module: cisco.mso.mso_schema
- module: cisco.mso.mso_schema_template
- module: cisco.mso.mso_schema_template_anp
- module: cisco.mso.mso_schema_template_anp_epg
- module: cisco.mso.mso_schema_template_contract_filter
extends_documentation_fragment: cisco.mso.modules
"""

EXAMPLES = r"""
- name: Bind an schema template contract with EPG Intra-EPG Contracts
  cisco.mso.mso_schema_template_anp_epg_intra_epg_contract:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: ansible_schema
    template: ansible_template
    anp: ansible_anp
    epg: ansible_epg
    contract:
      schema: ansible_contract_schema
      template: ansible_contract_template
      name: ansible_contract_name
    state: present

- name: Query an Intra-EPG Contract with contract details
  cisco.mso.mso_schema_template_anp_epg_intra_epg_contract:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: ansible_schema
    template: ansible_template
    anp: ansible_anp
    epg: ansible_epg
    contract:
      schema: ansible_contract_schema
      template: ansible_contract_template
      name: ansible_contract_name
    state: query
  register: query_with_contract_details

- name: Query an Intra-EPG Contract with only contract name
  cisco.mso.mso_schema_template_anp_epg_intra_epg_contract:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: ansible_schema
    template: ansible_template
    anp: ansible_anp
    epg: ansible_epg
    contract:
      name: ansible_contract_name
    state: query
  register: query_with_contract_name

- name: Query all Intra-EPG Contracts
  cisco.mso.mso_schema_template_anp_epg_intra_epg_contract:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: ansible_schema
    template: ansible_template
    anp: ansible_anp
    epg: ansible_epg
    state: query
  register: query_all

- name: Remove an Intra-EPG Contract with contract details
  cisco.mso.mso_schema_template_anp_epg_intra_epg_contract:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: ansible_schema
    template: ansible_template
    anp: ansible_anp
    epg: ansible_epg
    contract:
      schema: ansible_contract_schema
      template: ansible_contract_template
      name: ansible_contract_name
    state: absent

- name: Remove an Intra-EPG Contract with only contract name
  cisco.mso.mso_schema_template_anp_epg_intra_epg_contract:
    host: mso_host
    username: admin
    password: SomeSecretPassword
    schema: ansible_schema
    template: ansible_template
    anp: ansible_anp
    epg: ansible_epg
    contract:
      name: ansible_contract_name
    state: absent
"""

RETURN = r"""
"""


from ansible.module_utils.basic import AnsibleModule
from ansible_collections.cisco.mso.plugins.module_utils.mso import MSOModule, mso_argument_spec
from ansible_collections.cisco.mso.plugins.module_utils.schema import MSOSchema


def get_intra_epg_contracts_ref(intra_epg_contracts, name, schema_id, template):
    res_intra_epg_contracts_ref = []
    for intra_epg_contract in intra_epg_contracts:
        if intra_epg_contract.get("contractRef") == "/schemas/{0}/templates/{1}/contracts/{2}".format(schema_id, template, name):
            res_intra_epg_contracts_ref.append(intra_epg_contract.get("contractRef"))
        elif name is None:
            res_intra_epg_contracts_ref.append(intra_epg_contract.get("contractRef"))

    return res_intra_epg_contracts_ref


def main():
    argument_spec = mso_argument_spec()

    argument_spec.update(
        schema=dict(type="str", required=True),
        template=dict(type="str", required=True),
        anp=dict(type="str", aliases=["application_profile"], required=True),
        epg=dict(type="str", aliases=["endpoint_group"], required=True),
        contract=dict(
            type="dict",
            options=dict(
                name=dict(type="str", aliases=["contract_name"], required=True),
                schema=dict(type="str", aliases=["contract_schema"]),
                template=dict(type="str", aliases=["contract_template"]),
            ),
        ),
        state=dict(type="str", default="query", choices=["absent", "query", "present"]),
    )

    module = AnsibleModule(
        argument_spec=argument_spec,
        supports_check_mode=True,
        required_if=[
            ["state", "absent", ["contract"]],
            ["state", "present", ["contract"]],
        ],
    )

    schema = module.params.get("schema")
    template = str(module.params.get("template")).replace(" ", "")
    anp = module.params.get("anp")
    epg = module.params.get("epg")
    contract = module.params.get("contract") if module.params.get("contract") else {}
    state = module.params.get("state")

    mso = MSOModule(module)

    mso_schema = MSOSchema(mso, schema, template)
    mso_schema.set_template(template)
    mso_schema.set_template_anp(anp)
    mso_schema.set_template_anp_epg(epg)

    contract_template = (contract.get("template") if contract.get("template") else template).replace(" ", "")
    contract_schema_id = MSOSchema(mso, contract.get("schema"), contract_template).id if contract.get("schema") is not None else None

    matching_contracts = get_intra_epg_contracts_ref(
        mso_schema.schema_objects.get("template_anp_epg", {}).details.get("intraEpgContracts", []),
        contract.get("name"),
        schema_id=contract_schema_id if contract_schema_id else mso_schema.id,
        template=contract_template,
    )

    contract_ref = "/schemas/{0}/templates/{1}/contracts/{2}".format(
        contract_schema_id if contract_schema_id else mso_schema.id, contract_template, contract.get("name")
    )

    if contract_ref in matching_contracts:
        mso.existing = dict(contractRef=contract_ref)
        mso.previous = dict(contractRef=contract_ref)

    if state != "query":
        intra_epg_contract_path = "/templates/{0}/anps/{1}/epgs/{2}/intraEpgContracts/{3}".format(
            template, anp, epg, matching_contracts.index(contract_ref) if contract_ref in matching_contracts else "-"
        )
    else:
        mso.existing = [dict(contractRef=contract_ref) for contract_ref in matching_contracts]  # When the state is query

    ops = []
    if state == "present":
        mso_values = mso.existing if mso.existing else dict(contractRef=contract_ref)
        mso.sanitize(mso_values, collate=False, required=["contractRef"])
        ops.append(dict(op="replace" if mso.existing else "add", path=intra_epg_contract_path, value=mso_values))
    elif state == "absent":
        if mso.existing:
            ops.append(dict(op="remove", path=intra_epg_contract_path))

    if not module.check_mode and ops:
        # Returns No Contract - 204
        mso.request(mso_schema.path, method="PATCH", data=ops)

        mso_schema = MSOSchema(mso, schema, template)
        mso_schema.set_template(template)
        mso_schema.set_template_anp(anp)
        mso_schema.set_template_anp_epg(epg)

        matching_contracts = get_intra_epg_contracts_ref(
            mso_schema.schema_objects.get("template_anp_epg", {}).details.get("intraEpgContracts", []),
            contract.get("name"),
            schema_id=contract_schema_id if contract_schema_id else mso_schema.id,
            template=contract_template,
        )

        if matching_contracts and contract_ref in matching_contracts:
            mso.existing = dict(contractRef=contract_ref)  # When the state is present
        else:
            mso.existing = {}  # When the state is absent
    elif module.check_mode and state != "query":  # When the state is present/absent with check mode
        mso.existing = mso.proposed if state == "present" else {}

    mso.exit_json()


if __name__ == "__main__":
    main()



# (c) 2017, Ansible Project
#
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

from __future__ import absolute_import, division, print_function


__metaclass__ = type

from time import sleep

from ansible.module_utils.common.text.converters import to_bytes, to_text
from ansible.plugins.action import ActionBase
from ansible.utils.display import Display

from ansible_collections.ansible.netcommon.plugins.plugin_utils.compat import telnetlib


text_type = str
display = Display()


class ActionModule(ActionBase):
    TRANSFERS_FILES = False

    def run(self, tmp=None, task_vars=None):
        if self._task.environment and any(self._task.environment):
            self._display.warning("The telnet task does not support the environment keyword")

        result = super(ActionModule, self).run(tmp, task_vars)
        del tmp  # tmp no longer has any effect

        if self._play_context.check_mode:
            # in --check mode, always skip this module execution
            result["skipped"] = True
            result["msg"] = "The telnet task does not support check mode"
        else:
            result["changed"] = True
            result["failed"] = False

            host = to_text(self._task.args.get("host", self._play_context.remote_addr))
            user = to_text(self._task.args.get("user", self._play_context.remote_user))
            password = to_text(self._task.args.get("password", self._play_context.password))

            # FIXME, default to play_context?
            port = int(self._task.args.get("port", 23))
            timeout = int(self._task.args.get("timeout", 120))
            pause = int(self._task.args.get("pause", 1))

            send_newline = self._task.args.get("send_newline", False)
            crlf = self._task.args.get("crlf", False)

            login_prompt = to_text(self._task.args.get("login_prompt", "login: "))
            password_prompt = to_text(self._task.args.get("password_prompt", "Password: "))
            prompts = self._task.args.get("prompts", ["\\$ "])
            commands = self._task.args.get("command") or self._task.args.get("commands")

            if crlf:
                line_ending = "\r\n"
            else:
                line_ending = "\n"

            if isinstance(commands, text_type):
                commands = commands.split(",")

            if isinstance(commands, list) and commands:
                self.tn = telnetlib.Telnet(host, port, timeout)

                self.output = bytes()
                try:
                    if send_newline:
                        self.tn.write(to_bytes(line_ending))

                    self.await_prompts([login_prompt], timeout)
                    display.vvvvv(">>>user: %s" % user)
                    self.tn.write(to_bytes(user + line_ending))

                    if password:
                        self.await_prompts([password_prompt], timeout)
                        display.vvvvv(">>>password: %s" % password)
                        self.tn.write(to_bytes(password + line_ending))

                    self.await_prompts(prompts, timeout)

                    for cmd in commands:
                        display.vvvvv(">>> %s" % cmd)
                        self.tn.write(to_bytes(cmd + line_ending))
                        self.await_prompts(prompts, timeout)
                        display.vvvvv("<<< %s" % cmd)
                        sleep(pause)

                    self.tn.write(to_bytes("exit" + line_ending))

                except EOFError as e:
                    result["failed"] = True
                    result["msg"] = "Telnet action failed: %s" % to_text(e)
                except TimeoutError as e:
                    result["failed"] = True
                    result["msg"] = "Telnet timed out trying to find prompt(s): '%s'" % to_text(e)
                finally:
                    if self.tn:
                        self.tn.close()
                    result["stdout"] = to_text(self.output)
                    result["stdout_lines"] = self.output.splitlines(True)
            else:
                result["failed"] = True
                result["msg"] = "Telnet requires a command to execute"

        return result

    def await_prompts(self, prompts, timeout):
        index, match, out = self.tn.expect(list(map(to_bytes, prompts)), timeout=timeout)
        self.output += out
        if not match:
            raise TimeoutError(prompts)

        return index



#!/usr/bin/python
#
# This file is part of Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.
#
"""
The module file for ios_lldp_global
"""
from __future__ import absolute_import, division, print_function


__metaclass__ = type

DOCUMENTATION = """
module: ios_lldp_global
short_description: Resource module to configure LLDP.
description:
  This module configures and manages the Link Layer Discovery Protocol(LLDP)
  attributes on IOS platforms.
version_added: 1.0.0
author: Sumit Jaiswal (@justjais)
notes:
  - Tested against Cisco IOSXE Version 17.3 on CML.
  - This module works with connection C(network_cli).
    See U(https://docs.ansible.com/ansible/latest/network/user_guide/platform_ios.html)
options:
  config:
    description: A dictionary of LLDP options
    type: dict
    suboptions:
      holdtime:
        description:
          - LLDP holdtime (in sec) to be sent in packets.
          - Refer to vendor documentation for valid values.
        type: int
      reinit:
        description:
          - Specify the delay (in secs) for LLDP to initialize.
          - Refer to vendor documentation for valid values.
          - NOTE, if LLDP reinit is configured with a starting value, idempotency won't
            be maintained as the Cisco device doesn't record the starting reinit configured
            value. As such, Ansible cannot verify if the respective starting reinit
            value is already configured or not from the device side. If you try to apply
            starting reinit value in every play run, Ansible will show changed as True.
            For any other reinit value, idempotency will be maintained since any other
            reinit value is recorded in the Cisco device.
        type: int
      enabled:
        description:
          - Enable LLDP
        type: bool
      timer:
        description:
          - Specify the rate at which LLDP packets are sent (in sec).
          - Refer to vendor documentation for valid values.
        type: int
      tlv_select:
        description:
          - Selection of LLDP TLVs i.e. type-length-value to send
          - NOTE, if tlv-select is configured idempotency won't be maintained as Cisco
            device doesn't record configured tlv-select options. As such, Ansible cannot
            verify if the respective tlv-select options is already configured or not
            from the device side. If you try to apply tlv-select option in every play
            run, Ansible will show changed as True.
        type: dict
        suboptions:
          four_wire_power_management:
            description:
              - Cisco 4-wire Power via MDI TLV
            type: bool
          mac_phy_cfg:
            description:
              - IEEE 802.3 MAC/Phy Configuration/status TLV
            type: bool
          management_address:
            description:
              - Management Address TLV
            type: bool
          port_description:
            description:
              - Port Description TLV
            type: bool
          port_vlan:
            description:
              - Port VLAN ID TLV
            type: bool
          power_management:
            description:
              - IEEE 802.3 DTE Power via MDI TLV
            type: bool
          system_capabilities:
            description:
              - System Capabilities TLV
            type: bool
          system_description:
            description:
              - System Description TLV
            type: bool
          system_name:
            description:
              - System Name TLV
            type: bool
  running_config:
    description:
      - This option is used only with state I(parsed).
      - The value of this option should be the output received from the IOS device
        by executing the command B(show running-config | section ^lldp).
      - The state I(parsed) reads the configuration from C(running_config) option and
        transforms it into Ansible structured data as per the resource module's argspec
        and the value is then returned in the I(parsed) key within the result.
    type: str
  state:
    description:
      - The state the configuration should be left in
      - The module have declaratively similar behavior for replaced and overridden state.
      - The states I(rendered), I(gathered) and I(parsed) does not perform any change
        on the device.
      - The state I(rendered) will transform the configuration in C(config) option to
        platform specific CLI commands which will be returned in the I(rendered) key
        within the result. For state I(rendered) active connection to remote host is
        not required.
      - The state I(gathered) will fetch the running configuration from device and transform
        it into structured data in the format as per the resource module argspec and
        the value is returned in the I(gathered) key within the result.
      - The state I(parsed) reads the configuration from C(running_config) option and
        transforms it into JSON format as per the resource module parameters and the
        value is returned in the I(parsed) key within the result. The value of C(running_config)
        option should be the same format as the output of command I(show running-config
        | include ip route|ipv6 route) executed on device. For state I(parsed) active
        connection to remote host is not required.
    type: str
    choices:
      - merged
      - replaced
      - overridden
      - deleted
      - rendered
      - gathered
      - parsed
    default: merged
"""

EXAMPLES = """
# Using merged

# Before state:
# -------------
# vios#sh running-config | section ^lldp
# vios1#

- name: Merge provided configuration with device configuration
  cisco.ios.ios_lldp_global:
    config:
      holdtime: 10
      enabled: true
      reinit: 3
      timer: 10
    state: merged

# After state:
# ------------
# vios#sh running-config | section ^lldp
#  lldp timer 10
#  lldp holdtime 10
#  lldp reinit 3
#  lldp run

# Using replaced

# Before state:
# -------------
# vios#sh running-config | section ^lldp
#  lldp timer 10
#  lldp holdtime 10
#  lldp reinit 3
#  lldp run

- name: Replaces LLDP device configuration with provided configuration
  cisco.ios.ios_lldp_global:
    config:
      holdtime: 20
      reinit: 5
    state: replaced

# After state:
# -------------
# vios#sh running-config | section ^lldp
#  lldp holdtime 20
#  lldp reinit 5

# Using Deleted without any config passed
# "(NOTE: This will delete all of configured LLDP module attributes)"

# Before state:
# -------------
# vios#sh running-config | section ^lldp
#  lldp timer 10
#  lldp holdtime 10
#  lldp reinit 3
#  lldp run

- name: Delete LLDP attributes
  cisco.ios.ios_lldp_global:
    state: deleted

# After state:
# -------------
# vios#sh running-config | section ^lldp
# vios1#

# Using Gathered

# Before state:
# -------------
#
# vios#sh running-config | section ^lldp
#  lldp timer 10
#  lldp holdtime 10
#  lldp reinit 3
#  lldp run

- name: Gather listed interfaces with provided configurations
  cisco.ios.ios_lldp_global:
    config:
    state: gathered

# Module Execution Result:
# ------------------------
#
# "gathered": {
#         "enabled": true,
#         "holdtime": 10,
#         "reinit": 3,
#         "timer": 10
#     }

# After state:
# ------------
#
# vios#sh running-config | section ^lldp
#  lldp timer 10
#  lldp holdtime 10
#  lldp reinit 3
#  lldp run

# Using Rendered
- name: Render the commands for provided  configuration
  cisco.ios.ios_lldp_global:
    config:
      holdtime: 10
      enabled: true
      reinit: 3
      timer: 10
    state: rendered

# Module Execution Result:
# ------------------------
#
# "rendered": [
#         "lldp holdtime 10",
#         "lldp run",
#         "lldp timer 10",
#         "lldp reinit 3"
#     ]

# Using Parsed

# File: parsed.cfg
# ----------------
#
# lldp timer 10
# lldp holdtime 10
# lldp reinit 3
# lldp run

- name: Parse the commands for provided configuration
  cisco.ios.ios_lldp_global:
    running_config: "{{ lookup('file', 'parsed.cfg') }}"
    state: parsed

# Module Execution Result:
# ------------------------
#
# "parsed": {
#         "enabled": true,
#         "holdtime": 10,
#         "reinit": 3,
#         "timer": 10
#     }
"""

RETURN = """
before:
  description: The configuration as structured data prior to module invocation.
  returned: always
  type: dict
  sample: The configuration returned will always be in the same format of the parameters above.
after:
  description: The configuration as structured data after module completion.
  returned: when changed
  type: dict
  sample: The configuration returned will always be in the same format of the parameters above.
commands:
  description: The set of commands pushed to the remote device
  returned: always
  type: list
  sample: ['lldp holdtime 10', 'lldp run', 'lldp timer 10']
"""

from ansible.module_utils.basic import AnsibleModule

from ansible_collections.cisco.ios.plugins.module_utils.network.ios.argspec.lldp_global.lldp_global import (
    Lldp_globalArgs,
)
from ansible_collections.cisco.ios.plugins.module_utils.network.ios.config.lldp_global.lldp_global import (
    Lldp_global,
)


def main():
    """
    Main entry point for module execution

    :returns: the result form module invocation
    """
    required_if = [
        ("state", "merged", ("config",)),
        ("state", "replaced", ("config",)),
        ("state", "overridden", ("config",)),
        ("state", "rendered", ("config",)),
        ("state", "parsed", ("running_config",)),
    ]
    mutually_exclusive = [("config", "running_config")]

    module = AnsibleModule(
        argument_spec=Lldp_globalArgs.argument_spec,
        required_if=required_if,
        mutually_exclusive=mutually_exclusive,
        supports_check_mode=True,
    )
    result = Lldp_global(module).execute_module()
    module.exit_json(**result)


if __name__ == "__main__":
    main()



#!/usr/bin/python
# -*- coding: utf-8 -*-

# Copyright: (c) 2017, Sertac Ozercan <seozerca@microsoft.com>
# GNU General Public License v3.0+ (see COPYING or https://www.gnu.org/licenses/gpl-3.0.txt)

from __future__ import absolute_import, division, print_function
__metaclass__ = type

DOCUMENTATION = '''
---
module: azure_rm_virtualmachinescaleset_info

version_added: "0.1.2"

short_description: Get Virtual Machine Scale Set facts

description:
    - Get facts for a virtual machine scale set.
    - Note that this module was called M(azure.azcollection.azure_rm_virtualmachine_scaleset_facts) before Ansible 2.8. The usage did not change.

options:
    name:
        description:
            - Limit results to a specific virtual machine scale set.
        type: str
    resource_group:
        description:
            - The resource group to search for the desired virtual machine scale set.
        type: str
    tags:
        description:
            - Limit results by providing a list of tags. Format tags as 'key' or 'key:value'.
        type: list
        elements: str
    format:
        description:
            - Format of the data returned.
            - If C(raw) is selected information will be returned in raw format from Azure Python SDK.
            - If C(curated) is selected the structure will be identical to input parameters of M(azure.azcollection.azure_rm_virtualmachinescaleset) module.
            - In Ansible 2.5 and lower facts are always returned in raw format.
            - Please note that this option will be deprecated in 2.10 when curated format will become the only supported format.
        default: 'raw'
        type: str
        choices:
            - 'curated'
            - 'raw'

extends_documentation_fragment:
    - azure.azcollection.azure

author:
    - Sertac Ozercan (@sozercan)
'''

EXAMPLES = '''
- name: Get facts for a virtual machine scale set
  azure_rm_virtualmachinescaleset_info:
    resource_group: myResourceGroup
    name: testvmss001
    format: curated

- name: Get facts for all virtual networks
  azure_rm_virtualmachinescaleset_info:
    resource_group: myResourceGroup

- name: Get facts by tags
  azure_rm_virtualmachinescaleset_info:
    resource_group: myResourceGroup
    tags:
      - testing
'''

RETURN = '''
vmss:
    description:
        - List of virtual machine scale sets.
    returned: always
    type: list
    sample: [{
        "constrained_maximum_capacity": false,
        "etag": "3",
        "id": "/subscriptions/xxx-xxx/resourceGroups/v-xisuRG02/providers/Microsoft.Compute/virtualMachineScaleSets/testVMSStestvmss",
        "location": "eastus",
        "name": "testVMSStestvmss",
        "orchestration_mode": "Flexible",
        "platform_fault_domain_count": 1,
        "provisioning_state": "Succeeded",
        "single_placement_group": false,
        "sku": {
            "capacity": 1,
            "name": "Standard_A1_v2",
            "tier": "Standard"
        },
        "tags": {
            "key2": "value2",
            "key3": "value3"
        },
        "time_created": "2025-04-22T08:03:29.427822Z",
        "type": "Microsoft.Compute/virtualMachineScaleSets",
        "unique_id": "da4393f5-6060-4a0e-8ae5-7616316402b8",
        "upgrade_policy": {
            "mode": "Manual"
        },
        "virtual_machine_profile": {
            "network_profile": {
                "network_api_version": "2020-11-01",
                "network_interface_configurations": [
                    {
                        "auxiliary_mode": "None",
                        "auxiliary_sku": "None",
                        "delete_option": "Delete",
                        "disable_tcp_state_tracking": false,
                        "dns_settings": {
                            "dns_servers": []
                        },
                        "enable_ip_forwarding": false,
                        "ip_configurations": [
                            {
                                "application_gateway_backend_address_pools": [],
                                "application_security_groups": [
                                    {
                                        "id": "/subscriptions/xxx-xxx/resourceGroups/v-xisuRG02/providers/Microsoft.Network/applicationSecurityGroups/apptestvmss02"
                                    },
                                    {
                                        "id": "/subscriptions/xxx-xxx/resourceGroups/v-xisuRG02/providers/Microsoft.Network/applicationSecurityGroups/apptestvmss"
                                    }
                                ],
                                "load_balancer_backend_address_pools": [],
                                "name": "default",
                                "primary": true,
                                "private_ip_address_version": "IPv4",
                                "public_ip_address_configuration": {
                                    "idle_timeout_in_minutes": 4,
                                    "ip_tags": [],
                                    "name": "instancepublicip",
                                    "public_ip_address_version": "IPv4"
                                },
                                "subnet": {
                                    "id": "/subscriptions/xxx-xxx/resourceGroups/v-xisuRG02/providers/Microsoft.Network/virtualNetworks/VMSStestVnet/subnets/VMSStestSubnet"
                                }
                            }
                        ],
                        "name": "testVMSStestvmss",
                        "primary": true
                    }
                ]
            },
            "os_profile": {
                "admin_username": "testuser",
                "allow_extension_operations": true,
                "computer_name_prefix": "testVMSStestvmss",
                "linux_configuration": {
                    "disable_password_authentication": true,
                    "patch_settings": {
                        "assessment_mode": "ImageDefault",
                        "patch_mode": "ImageDefault"
                    },
                    "provision_vm_agent": true,
                    "ssh": {
                        "public_keys": [
                            {
                                "key_data": "ssh-rsa xxxxxxxxxx xx.yy@qq.com",
                                "path": "/home/testuser/.ssh/authorized_keys"
                            }
                        ]
                    }
                },
                "require_guest_provision_signal": true,
                "secrets": []
            },
            "storage_profile": {
                "data_disks": [
                    {
                        "caching": "ReadWrite",
                        "create_option": "Empty",
                        "delete_option": "Delete",
                        "disk_size_gb": 64,
                        "lun": 0,
                        "managed_disk": {
                            "storage_account_type": "Standard_LRS"
                        }
                    }
                ],
                "image_reference": {
                    "offer": "0001-com-ubuntu-server-focal",
                    "publisher": "Canonical",
                    "sku": "20_04-lts",
                    "version": "20.04.202504030"
                },
                "os_disk": {
                    "caching": "ReadWrite",
                    "create_option": "FromImage",
                    "delete_option": "Delete",
                    "disk_size_gb": 30,
                    "managed_disk": {
                        "storage_account_type": "Standard_LRS"
                    },
                    "os_type": "Linux"
                }
            },
            "time_created": "2025-04-22T08:53:17.201851Z"
        }
    }]
'''  # NOQA

from ansible_collections.azure.azcollection.plugins.module_utils.azure_rm_common import AzureRMModuleBase
import re

try:
    from azure.core.exceptions import ResourceNotFoundError
except Exception:
    # handled in azure_rm_common
    pass

AZURE_OBJECT_CLASS = 'VirtualMachineScaleSet'

AZURE_ENUM_MODULES = ['azure.mgmt.compute.models']


class AzureRMVirtualMachineScaleSetInfo(AzureRMModuleBase):
    """Utility class to get virtual machine scale set facts"""

    def __init__(self):

        self.module_args = dict(
            name=dict(type='str'),
            resource_group=dict(type='str'),
            tags=dict(type='list', elements='str'),
            format=dict(
                type='str',
                choices=['curated',
                         'raw'],
                default='raw'
            )
        )

        self.results = dict(
            changed=False,
        )

        self.name = None
        self.resource_group = None
        self.format = None
        self.tags = None

        super(AzureRMVirtualMachineScaleSetInfo, self).__init__(
            derived_arg_spec=self.module_args,
            supports_check_mode=True,
            supports_tags=False,
            facts_module=True
        )

    def exec_module(self, **kwargs):

        for key in self.module_args:
            setattr(self, key, kwargs[key])

        if self.name and not self.resource_group:
            self.fail("Parameter error: resource group required when filtering by name.")

        if self.name:
            result = self.get_item()
        else:
            result = self.list_items()

        if self.format == 'curated':
            for index in range(len(result)):
                vmss = result[index]
                subnet_name = None
                load_balancer_name = None
                virtual_network_name = None
                ssh_password_enabled = False

                try:
                    subnet_id = (vmss['virtual_machine_profile']['network_profile']['network_interface_configurations'][0]
                                 ['ipConfigurations'][0]['subnet']['id'])
                    subnet_name = re.sub('.*subnets\\/', '', subnet_id)
                except Exception:
                    self.log('Could not extract subnet name')

                try:
                    backend_address_pool_id = (vmss['virtual_machine_profile']['network_profile']['network_interface_configurations'][0]
                                               ['ip_configurations'][0]['load_balancer_backend_address_pools'][0]['id'])
                    load_balancer_name = re.sub('\\/backendAddressPools.*', '', re.sub('.*loadBalancers\\/', '', backend_address_pool_id))
                    virtual_network_name = re.sub('.*virtualNetworks\\/', '', re.sub('\\/subnets.*', '', subnet_id))
                except Exception:
                    self.log('Could not extract load balancer / virtual network name')

                try:
                    ssh_password_enabled = (not vmss['virtual_machine_profile']['os_profile']
                                                    ['linux_configuration']['disable_password_authentication'])
                except Exception:
                    self.log('Could not extract SSH password enabled')

                data_disks = vmss['virtual_machine_profile']['storage_profile'].get('data_disks', [])

                for disk_index in range(len(data_disks)):
                    old_disk = data_disks[disk_index]
                    new_disk = {
                        'lun': old_disk['lun'],
                        'disk_size_gb': old_disk['disk_size_gb'],
                        'managed_disk_type': old_disk['managed_disk']['storage_account_type'],
                        'caching': old_disk['caching']
                    }
                    data_disks[disk_index] = new_disk

                updated = {
                    'id': vmss['id'],
                    'resource_group': self.resource_group,
                    'name': vmss['name'],
                    'state': 'present',
                    'location': vmss['location'],
                    'vm_size': vmss['sku']['name'],
                    'capacity': vmss['sku']['capacity'],
                    'tier': vmss['sku']['tier'],
                    'upgrade_policy': vmss.get('upgrade_policy'),
                    'orchestrationMode': vmss.get('orchestration_mode'),
                    'platformFaultDomainCount': vmss.get('platform_fault_domain_count'),
                    'admin_username': vmss['virtual_machine_profile']['os_profile']['admin_username'],
                    'admin_password': vmss['virtual_machine_profile']['os_profile'].get('admin_password'),
                    'ssh_password_enabled': ssh_password_enabled,
                    'image': vmss['virtual_machine_profile']['storage_profile']['image_reference'],
                    'os_disk_caching': vmss['virtual_machine_profile']['storage_profile']['os_disk']['caching'],
                    'os_type': 'Linux' if (vmss['virtual_machine_profile']['os_profile'].get('linux_configuration') is not None) else 'Windows',
                    'overprovision': vmss.get('overprovision'),
                    'managed_disk_type': vmss['virtual_machine_profile']['storage_profile']['os_disk']['managed_disk']['storage_account_type'],
                    'data_disks': data_disks,
                    'virtual_network_name': virtual_network_name,
                    'subnet_name': subnet_name,
                    'load_balancer': load_balancer_name,
                    'identity': vmss.get('identity', None),
                    'tags': vmss.get('tags')
                }

                result[index] = updated

        self.results['vmss'] = result

        return self.results

    def get_item(self):
        """Get a single virtual machine scale set"""

        self.log('Get properties for {0}'.format(self.name))

        item = None
        results = []

        try:
            item = self.compute_client.virtual_machine_scale_sets.get(self.resource_group, self.name)
        except ResourceNotFoundError:
            pass

        if item and self.has_tags(item.tags, self.tags):
            results = [self.serialize_obj(item, AZURE_OBJECT_CLASS, enum_modules=AZURE_ENUM_MODULES)]

        return results

    def list_items(self):
        """Get all virtual machine scale sets"""

        self.log('List all virtual machine scale sets')

        try:
            response = self.compute_client.virtual_machine_scale_sets.list(self.resource_group)
        except ResourceNotFoundError as exc:
            self.fail('Failed to list all items - {0}'.format(str(exc)))

        results = []
        for item in response:
            if self.has_tags(item.tags, self.tags):
                results.append(self.serialize_obj(item, AZURE_OBJECT_CLASS, enum_modules=AZURE_ENUM_MODULES))

        return results


def main():
    """Main module execution code path"""

    AzureRMVirtualMachineScaleSetInfo()


if __name__ == '__main__':
    main()
